"""Unit test for cosmic
"""

__author__ = 'Katie Breivik <katie.breivik@gmail.com>'

import os
import unittest
import numpy as np
import pandas as pd
from ..sample import InitialBinaryTable
from ..sample.sampler.independent import Sample
from ..sample.sampler.multidim import MultiDim
from ..sample.sampler.cmc import CMCSample
from ..sample.cmc import elson
from ..sample.initialcmctable import InitialCMCTable
from scipy.optimize import curve_fit
from ..utils import a_from_p, get_porb_norm

SAMPLECLASS = Sample()
MULTIDIMSAMPLECLASS = MultiDim()
CMCSAMPLECLASS = CMCSample()
TEST_DATA_DIR = os.path.join(os.path.split(__file__)[0], 'data')

#distribution slopes/power laws
KROUPA_93_HI = -2.7
KROUPA_93_MID = -2.2
KROUPA_93_LO = -1.3
KROUPA_01_HI = -2.3
KROUPA_01_LO = -1.3
SALPETER_55 = -2.35
SANA12_PORB_POWER_LAW = -0.55
FLAT_SLOPE = 0.0
# Since we integrate up to 0.7 in ecc, the slope is 4.0 rounded at the 0th decimal
THERMAL_SLOPE = 4.0
SANA12_ECC_POWER_LAW = -0.45
SANA12_ECC_POWER_LAW_ROUND = -0.5
MEAN_RAGHAVAN = 4.9
SIGMA_RAGHAVAN = 2.3
CLOSE_BINARY_FRAC = 0.42


N_BINARY_SELECT = 85
VANHAAFTEN_BINFRAC_MAX = 0.9989087986493874
VANHAAFTEN_BINFRAC_MIN = 0.6192803136799157
MULTIDIM_BINFRAC_MAX = 0.6146916774140262
MULTIDIM_BINFRAC_MIN = 0.13786300908773025
CONST_SFR_SUM = 460028.2453521937
BURST_SFR_SUM = 946002.8245352195
KSTAR_SOLAR = 1.0
MOE_TOTAL_MASS = 20.27926225850954
METALLICITY_1000 = 0.02
METALLICITY_13000 = 0.02*0.15

QMIN_LOWMASS = 0.1
QMIN_HIGHMASS = 0.7
M2MIN_LOWMASS = 0.08
M2MIN_HIGHMASS = 12.0
BINFRAC_LOWMASS = 499139
BINFRAC_HIGHMASS = 2205

KING_TEST_DATA = np.load(os.path.join(TEST_DATA_DIR, "cmc_king_test.npz"))
ELSON_TEST_DATA = np.load(os.path.join(TEST_DATA_DIR, "cmc_elson_test.npz"))
PLUMMER_TEST_DATA = np.load(os.path.join(TEST_DATA_DIR, "cmc_plummer_test.npz"))
R_PLUMMER_TEST_ARRAY, VR_PLUMMER_TEST_ARRAY, VT_PLUMMER_TEST_ARRAY = PLUMMER_TEST_DATA["arr_0"], PLUMMER_TEST_DATA["arr_1"], PLUMMER_TEST_DATA["arr_2"] 
R_ELSON_TEST_ARRAY, VR_ELSON_TEST_ARRAY, VT_ELSON_TEST_ARRAY = ELSON_TEST_DATA["arr_0"], ELSON_TEST_DATA["arr_1"], ELSON_TEST_DATA["arr_2"] 
R_KING_TEST_ARRAY, VR_KING_TEST_ARRAY, VT_KING_TEST_ARRAY = KING_TEST_DATA["arr_0"], KING_TEST_DATA["arr_1"], KING_TEST_DATA["arr_2"]

REFF_TEST_ARRAY = np.array([3.94190562, 5.99895482])

SINGLES_CMC_FITS, BINARIES_CMC_FITS = InitialCMCTable.read(filename=os.path.join(TEST_DATA_DIR, "input_cmc.fits"))
SINGLES_CMC_HDF5, BINARIES_CMC_HDF5 = InitialCMCTable.read(filename=os.path.join(TEST_DATA_DIR, "input_cmc.hdf5"))

def power_law_fit(data, n_bins=100):
    def line(x, a, b):
        return x*a + b
    def center_bins(bins):
        mid_bin = []
        for bin_lo, bin_hi in zip(bins[:-1], bins[1:]):
            mid_bin.append(bin_lo + (bin_hi-bin_lo)/2)
        return mid_bin

    hist, bins = np.histogram(data, bins=n_bins)
    bins = center_bins(bins)

    xdata = np.log10(bins)
    ydata = np.log10(hist)

    popt, pcov = curve_fit(line, xdata, ydata)

    slope, intercept = popt[0], popt[1]

    return slope

def linear_fit(data):
    def line(x, a, b):
        return x*a + b
    def center_bins(bins):
        mid_bin = []
        for bin_lo, bin_hi in zip(bins[:-1], bins[1:]):
            mid_bin.append(bin_lo + (bin_hi-bin_lo)/2)
        return mid_bin

    hist, bins = np.histogram(data, bins=100, density=True)
    bins = center_bins(bins)
    popt, pcov = curve_fit(line, bins, hist)

    slope, intercept = popt[0], popt[1]

    return slope

class TestSample(unittest.TestCase):
    """`TestCase` for the cosmic Sample class, which generates several
        independent initial parameters drawn from specified distributions
    """

    def test_sample_primary_kroupa93(self):
        np.random.seed(2)
        # Check that the sample_primary function samples mass correctly
        mass, total_mass = SAMPLECLASS.sample_primary(primary_model='kroupa93', size=1000000)
        mass_hi = mass[mass > 1.0]
        # filter out highest masses because they kill us with the histograms
        mass_hi = mass_hi[mass_hi < 10.0]
        mass_mid = mass[(mass <= 1.0)]
        mass_mid = mass_mid[mass_mid > 0.5]
        mass_lo = mass[mass <= 0.5]

        #few bins for the most massive stars
        power_slope_hi = power_law_fit(mass_hi, n_bins=50)
        power_slope_mid = power_law_fit(mass_mid)
        power_slope_lo = power_law_fit(mass_lo)

        self.assertEqual(np.round(power_slope_hi, 1), KROUPA_93_HI)
        self.assertEqual(np.round(power_slope_mid, 1), KROUPA_93_MID)
        self.assertEqual(np.round(power_slope_lo, 1), KROUPA_93_LO)


    def test_sample_primary_kroupa01(self):
        np.random.seed(2)
        # Check that the sample_primary function samples mass correctly
        mass, total_mass = SAMPLECLASS.sample_primary(primary_model='kroupa01', size=1000000)
        mass_hi = mass[mass > 0.5]
        # filter out highest masses because they kill us with the histograms
        mass_hi = mass_hi[mass_hi < 10.0]
        mass_lo = mass[mass <= 0.5]

        #few bins for the most massive stars
        power_slope_hi = power_law_fit(mass_hi, n_bins=50)
        power_slope_lo = power_law_fit(mass_lo)

        self.assertEqual(np.round(power_slope_hi, 1), KROUPA_01_HI)
        self.assertEqual(np.round(power_slope_lo, 1), KROUPA_01_LO)

    def test_sample_primary_salpeter55(self):
        np.random.seed(3)
        # Check that the sample_primary function samples mass correctly
        mass, total_mass = SAMPLECLASS.sample_primary(primary_model='salpeter55', size=10000000)
        #filter on mass to get better statistics
        power_slope = power_law_fit(mass[mass < 1.0], n_bins=50)
        self.assertEqual(np.round(power_slope, 2), SALPETER_55)

    def test_sample_secondary(self):
        np.random.seed(2)
        # Check that the sample_secondary function samples secondary mass correctly
        mass1, total_mass = SAMPLECLASS.sample_primary(primary_model='salpeter55', size=10000000)
        ind_massive, = np.where(mass1 > 15.0)
        ind_not_massive, = np.where(mass1 < 15.0)

        np.random.seed(2)
        mass2 = SAMPLECLASS.sample_secondary(primary_mass=mass1, qmin=0.1)
        q = mass2[ind_massive] / mass1[ind_massive]
        slope = linear_fit(q)
        self.assertEqual(np.round(slope, 1), FLAT_SLOPE)

        np.random.seed(2)
        mass2 = SAMPLECLASS.sample_secondary(primary_mass=mass1, qmin=-1)
        q = mass2[ind_not_massive] / mass1[ind_not_massive]
        slope = linear_fit(q)
        self.assertEqual(np.round(slope, 1), FLAT_SLOPE)

        np.random.seed(2)
        mass2 = SAMPLECLASS.sample_secondary(primary_mass=mass1, m2_min=0.08)
        q = mass2[ind_massive] / mass1[ind_massive]
        slope = linear_fit(q)
        self.assertEqual(np.round(slope, 1), FLAT_SLOPE)

        np.random.seed(2)
        mass2 = SAMPLECLASS.sample_secondary(primary_mass=mass1, qmin=0.1, m2_min=0.08)
        q = mass2[ind_massive] / mass1[ind_massive]
        slope = linear_fit(q)
        self.assertEqual(np.round(slope, 1), FLAT_SLOPE)

    def test_binary_select(self):
        np.random.seed(2)
        # Check that the binary select function chooses binarity properly
        m1_b, m1_s, binfrac, bin_index = SAMPLECLASS.binary_select(primary_mass=np.arange(1,100), binfrac_model=0.0)
        self.assertEqual(len(m1_b), 0)
        m1_b, m1_s, binfrac, bin_index = SAMPLECLASS.binary_select(primary_mass=np.arange(1,100), binfrac_model=1.0)
        self.assertEqual(len(m1_b), 99)
        m1_b, m1_s, binfrac, bin_index = SAMPLECLASS.binary_select(primary_mass=np.arange(1,100), binfrac_model='vanHaaften')
        self.assertEqual(len(m1_b), N_BINARY_SELECT)

    def test_binary_fraction(self):
        np.random.seed(2)
        # Check that the binary fraction tracking is correct
        m1_b, m1_s, binfrac, bin_index = SAMPLECLASS.binary_select(primary_mass=np.arange(1,100), binfrac_model=0.5)
        self.assertEqual(binfrac.max(), 0.5)
        m1_b, m1_s, binfrac, bin_index = SAMPLECLASS.binary_select(primary_mass=np.arange(1,100), binfrac_model='vanHaaften')
        self.assertEqual(binfrac.max(), VANHAAFTEN_BINFRAC_MAX)
        self.assertEqual(binfrac.min(), VANHAAFTEN_BINFRAC_MIN)

    def test_msort(self):
        np.random.seed(2)
        mass1, total_mass = SAMPLECLASS.sample_primary(primary_model='kroupa01', size=1000000)
        # Check that qmin_msort and m2_min_msort are workings as expected
        mass2 = SAMPLECLASS.sample_secondary(primary_mass = mass1, qmin=0.1, m2_min=0.08, msort=15, qmin_msort=0.7, m2_min_msort=12)
        ind_light, = np.where(mass1 < 15.0)
        ind_massive, = np.where(mass1 >= 15.0)
        m2_light = mass2[ind_light]
        m2_massive = mass2[ind_massive]
        q_light = mass2[ind_light]/mass1[ind_light]
        q_massive = mass2[ind_massive]/mass1[ind_massive]
        assert m2_light.min() > M2MIN_LOWMASS
        assert m2_massive.min() > M2MIN_HIGHMASS
        assert q_light.min() > QMIN_LOWMASS
        assert q_massive.min() > QMIN_HIGHMASS
        # Check that the binary fraction tracking is correct when using msort
        m1_b, m1_s, binfrac, bin_index = SAMPLECLASS.binary_select(primary_mass=mass1, binfrac_model=0.5, msort=15, binfrac_model_msort=0.8)
        assert np.sum(binfrac==0.5) == BINFRAC_LOWMASS
        assert np.sum(binfrac==0.8) == BINFRAC_HIGHMASS

#
    def test_sample_porb(self):
        # next do Sana12
        np.random.seed(4)
        mass1, total_mass = SAMPLECLASS.sample_primary(primary_model='kroupa01', size=100000)
        mass2 = SAMPLECLASS.sample_secondary(primary_mass = mass1, qmin=0.1)
        rad1 = SAMPLECLASS.set_reff(mass=mass1, metallicity=0.02)
        rad2 = SAMPLECLASS.set_reff(mass=mass2, metallicity=0.02)
        porb,aRL_over_a = SAMPLECLASS.sample_porb(
            mass1, mass2, rad1, rad2, 'sana12', size=mass1.size
        )
        power_slope = power_law_fit(np.log10(porb))
        self.assertEqual(np.round(power_slope, 2), SANA12_PORB_POWER_LAW)

        np.random.seed(5)
        # next do Renzo+19
        m1_high = mass1+15
        rad1_high = SAMPLECLASS.set_reff(mass=m1_high, metallicity=0.02)
        porb,aRL_over_a = SAMPLECLASS.sample_porb(
            m1_high, mass2, rad1_high, rad2, 'renzo19', size=m1_high.size
        )
        porb_cut = porb[porb > 2.5]
        power_slope = power_law_fit(np.log10(porb_cut))
        self.assertAlmostEqual(np.round(power_slope, 2), SANA12_PORB_POWER_LAW)

        porb,aRL_over_a = SAMPLECLASS.sample_porb(
            mass1, mass2, rad1, rad2, 'renzo19', size=mass1.size
        )
        ind_log_uniform, = np.where(mass1 <= 15)
        porb_log_uniform = porb[ind_log_uniform]
        uniform_slope = linear_fit(np.log10(porb_log_uniform))
        self.assertEqual(np.round(uniform_slope, 1), FLAT_SLOPE)

        # next check the log uniform
        porb,aRL_over_a = SAMPLECLASS.sample_porb(
            mass1, mass2, rad1, rad2, 'log_uniform', size=mass1.size
        )
        power_slope = linear_fit(np.log10(porb))
        sep = a_from_p(porb, mass1, mass2)
        sep = sep[sep > 10]
        uniform_slope = linear_fit(np.log10(sep))
        self.assertEqual(np.round(uniform_slope, 1), FLAT_SLOPE)

        # next check raghavan10
        porb,aRL_over_a = SAMPLECLASS.sample_porb(
            mass1, mass2, rad1, rad2, 'raghavan10', size=mass1.size
        )
        log_porb_mean = np.mean(np.log10(porb))
        log_porb_sigma = np.std(np.log10(porb))
        self.assertTrue(np.round(log_porb_mean, 1) >= MEAN_RAGHAVAN-0.15)
        self.assertEqual(np.round(log_porb_sigma, 0), np.round(SIGMA_RAGHAVAN, 0))

        # next check moe19
        from ..utils import get_met_dep_binfrac
        from scipy.interpolate import interp1d
        from scipy.stats import kstest
        metallicity = 0.001
        # this is a metallicity dependent population:
        binfrac = get_met_dep_binfrac(metallicity)
        print(binfrac)
        mass1, total_mass = SAMPLECLASS.sample_primary(primary_model='kroupa01', size=100000)
        (mass1_binaries, mass_single, binfrac_binaries, binary_index,) = SAMPLECLASS.binary_select(
            mass1, binfrac_model=binfrac,
            )
        mass2_binaries = SAMPLECLASS.sample_secondary(
            primary_mass=mass1_binaries, qmin=0.1
            )
        rad1 = SAMPLECLASS.set_reff(mass=mass1_binaries, metallicity=metallicity)
        rad2 = SAMPLECLASS.set_reff(mass=mass2_binaries, metallicity=metallicity)
        
        
        porb,aRL_over_a = SAMPLECLASS.sample_porb(
            mass1_binaries, mass2_binaries, rad1, rad2, 'moe19', size=mass1_binaries.size, met=metallicity
        )
        
        binary_frac = len(porb) / (len(mass1))
        self.assertAlmostEqual(np.round(binary_frac, 2), binfrac)
        
                

    def test_sample_ecc(self):
        np.random.seed(2)
        # Check that the sample_ecc function samples ecc properly

        # first sample orbital periods
        np.random.seed(4)
        mass1, total_mass = SAMPLECLASS.sample_primary(primary_model='kroupa01', size=100000)
        mass2 = SAMPLECLASS.sample_secondary(primary_mass = mass1, qmin=0.1)
        rad1 = SAMPLECLASS.set_reff(mass=mass1, metallicity=0.02)
        rad2 = SAMPLECLASS.set_reff(mass=mass2, metallicity=0.02)
        porb,aRL_over_a = SAMPLECLASS.sample_porb(
            mass1, mass2, rad1, rad2, 'sana12', size=mass1.size
        )

        # now we feed aRL_over_a into sample_ecc
        ecc = SAMPLECLASS.sample_ecc(aRL_over_a, ecc_model='thermal', size=mass1.size)
        ecc_cut = ecc[ecc < 0.7]
        slope = linear_fit(ecc_cut)
        self.assertEqual(np.round(slope, 0), THERMAL_SLOPE)

        ecc = SAMPLECLASS.sample_ecc(aRL_over_a, ecc_model='sana12', size=mass1.size)
        ecc_cut = ecc[ecc < 0.91]
        power_slope = power_law_fit(ecc_cut)
        self.assertEqual(np.round(power_slope, 1), SANA12_ECC_POWER_LAW_ROUND)

        ecc = SAMPLECLASS.sample_ecc(aRL_over_a, ecc_model='circular', size=mass1.size)
        self.assertEqual(np.mean(ecc), 0.0)


    def test_sample_SFH(self):
        np.random.seed(2)
        # Check that the sample SFH function samples SFH='const' correctly
        times, met = SAMPLECLASS.sample_SFH(SF_start=10000.0,\
                                            SF_duration=10000.0,\
                                            met = 0.02, size=100)
        self.assertEqual(times.sum(), CONST_SFR_SUM)
        self.assertAlmostEqual(np.mean(met), 0.02)

        np.random.seed(2)
        # Check that the sample SFH function samples SFH='burst' correctly
        times, met = SAMPLECLASS.sample_SFH(SF_start=10000.0,\
                                            SF_duration=1000.0,\
                                            met = 0.02, size=100)
        self.assertEqual(times.sum(), BURST_SFR_SUM)
        self.assertAlmostEqual(np.mean(met), 0.02)

        # Check that the sample SFH function samples SFH='delta_burst' correctly
        times, met = SAMPLECLASS.sample_SFH(SF_start=10000.0,\
                                            SF_duration=0.0,\
                                            met = 0.02, size=100)
        self.assertEqual(times.sum(), 100*10000.0)
        self.assertAlmostEqual(np.mean(met), 0.02)

    def test_set_kstar(self):
        # Check that the kstar is selected properly
        kstar = SAMPLECLASS.set_kstar(pd.DataFrame([1.0, 1.0, 1.0, 1.0, 1.0]))
        self.assertEqual(np.mean(kstar), KSTAR_SOLAR)

    def test_Moe_sample(self):
        # Test the multidim sampler and system-by-system binary fraction
        m1, m2, porb, ecc, mass_singles, mass_binaries, n_singles, n_binaries, binfrac = MULTIDIMSAMPLECLASS.initial_sample(rand_seed = 2, size=10, nproc=1, mp_seeds=[0])
        self.assertEqual(np.sum(mass_singles), MOE_TOTAL_MASS)
        self.assertAlmostEqual(binfrac.max(), MULTIDIM_BINFRAC_MAX)
        self.assertAlmostEqual(binfrac.min(), MULTIDIM_BINFRAC_MIN)

    def test_sample_MultiDim_SFH(self):
        np.random.seed(2)
        # Check that the sample SFH function samples SFH='const' correctly
        times, met = MULTIDIMSAMPLECLASS.sample_SFH(SF_start=10000.0,\
                                                    SF_duration=10000.0,\
                                                    met = 0.02, size=100)
        self.assertEqual(times.sum(), CONST_SFR_SUM)
        self.assertAlmostEqual(np.mean(met), 0.02)

        np.random.seed(2)
        # Check that the sample SFH function samples SFH='burst' correctly
        times, met = MULTIDIMSAMPLECLASS.sample_SFH(SF_start=10000.0,\
                                                    SF_duration=1000.0,\
                                                    met = 0.02, size=100)
        self.assertEqual(times.sum(), BURST_SFR_SUM)
        self.assertAlmostEqual(np.mean(met), 0.02)

        # Check that the sample SFH function samples SFH='delta_burst' correctly
        times, met = MULTIDIMSAMPLECLASS.sample_SFH(SF_start=10000.0,\
                                                    SF_duration=0.0,\
                                                    met = 0.02, size=100)
        self.assertEqual(times.sum(), 100*10000.0)
        self.assertAlmostEqual(np.mean(met), 0.02)

    def test_set_kstar_MultiDim(self):
        # Check that the kstar is selected properly
        kstar = MULTIDIMSAMPLECLASS.set_kstar(pd.DataFrame([1.0, 1.0, 1.0, 1.0, 1.0]))
        self.assertEqual(np.mean(kstar), KSTAR_SOLAR)

    def test_sampling_binfrac_zero(self):
        # check that you can't sample based on size with a binary fraction of 0.0
        it_fails = False
        try:
            InitialBinaryTable.sampler('independent', np.arange(16), np.arange(16),
                                       primary_model='kroupa01', ecc_model='thermal',
                                       porb_model='sana12', binfrac_model=0.0,
                                       SF_start=10.0, SF_duration=0.0, met=0.02,
                                       size=1000)
        except ValueError:
            it_fails = True
        self.assertTrue(it_fails)

    def test_sampling_targets_bad_input(self):
        # check that you get an error if you don't supply a size or total_mass
        it_fails = False
        try:
            InitialBinaryTable.sampler('independent', np.arange(16), np.arange(16),
                                       primary_model='kroupa01', ecc_model='thermal',
                                       porb_model='sana12', binfrac_model=0.5,
                                       SF_start=10.0, SF_duration=0.0, met=0.02,
                                       sampling_target="total_mass")
        except ValueError:
            it_fails = True
        self.assertTrue(it_fails)
    
        it_fails = False
        try:
            InitialBinaryTable.sampler('independent', np.arange(16), np.arange(16),
                                       primary_model='kroupa01', ecc_model='thermal',
                                       porb_model='sana12', binfrac_model=0.5,
                                       SF_start=10.0, SF_duration=0.0, met=0.02,
                                       total_mass=None, sampling_target="total_mass")
        except ValueError:
            it_fails = True
        self.assertTrue(it_fails)
    
        it_fails = False
        try:
            InitialBinaryTable.sampler('independent', np.arange(16), np.arange(16),
                                       primary_model='kroupa01', ecc_model='thermal',
                                       porb_model='sana12', binfrac_model=0.5,
                                       SF_start=10.0, SF_duration=0.0, met=0.02,
                                       size=None, total_mass=None, sampling_target="size")
        except ValueError:
            it_fails = True
        self.assertTrue(it_fails)

    def test_sampling_targets_size(self):
        # check that you can sample based on size
        for size in np.random.randint(100, 1000, size=100):
            samples = InitialBinaryTable.sampler('independent', np.arange(16), np.arange(16),
                                                primary_model='kroupa01', ecc_model='thermal',
                                                porb_model='sana12', binfrac_model=0.5,
                                                SF_start=10.0, SF_duration=0.0, met=0.02,
                                                size=size)
            self.assertGreaterEqual(len(samples[0]), size)

    def test_sampling_targets_mass(self):
        # check that you can sample based on total mass
        for mass in np.random.randint(100, 1000, size=100):
            samples = InitialBinaryTable.sampler('independent', np.arange(16), np.arange(16),
                                                primary_model='kroupa01', ecc_model='thermal',
                                                porb_model='sana12', binfrac_model=0.5,
                                                SF_start=10.0, SF_duration=0.0, met=0.02,
                                                size=None, total_mass=mass, sampling_target="total_mass")
            self.assertGreaterEqual(samples[1] + samples[2], mass)

    def test_sampling_targets_mass_trimmed(self):
        # check that you can sample based on total mass and that the samples are trimmed properly
        for mass in np.random.randint(100, 1000, size=100):
            samples = InitialBinaryTable.sampler('independent', np.arange(16), np.arange(16),
                                                primary_model='kroupa01', ecc_model='thermal',
                                                porb_model='sana12', binfrac_model=0.5,
                                                SF_start=10.0, SF_duration=0.0, met=0.02,
                                                size=None, total_mass=mass, sampling_target="total_mass",
                                                trim_extra_samples=True)
            self.assertLessEqual(abs(samples[1] + samples[2] - mass), 300)

class TestCMCSample(unittest.TestCase):
    def test_plummer_profile(self):
        np.random.seed(2)
        r, vr, vt = CMCSAMPLECLASS.set_r_vr_vt('plummer',N=100, r_max=300)
        np.testing.assert_allclose(VR_PLUMMER_TEST_ARRAY, vr, rtol=1e-5)
        np.testing.assert_allclose(VT_PLUMMER_TEST_ARRAY, vt, rtol=1e-5)
        np.testing.assert_allclose(R_PLUMMER_TEST_ARRAY, r, rtol=1e-5)

    def test_elson_profile(self):
        np.random.seed(2)
        r, vr, vt = CMCSAMPLECLASS.set_r_vr_vt('elson',N=100, r_max=300, gamma=3)
        np.testing.assert_allclose(VR_ELSON_TEST_ARRAY, vr, rtol=1e-5)
        np.testing.assert_allclose(VT_ELSON_TEST_ARRAY, vt, rtol=1e-5)
        np.testing.assert_allclose(R_ELSON_TEST_ARRAY, r, rtol=1e-5)

    def test_king_profile(self):
        np.random.seed(2)
        r, vr, vt = CMCSAMPLECLASS.set_r_vr_vt('king',N=100, w_0=5)
        np.testing.assert_allclose(VR_KING_TEST_ARRAY, vr, rtol=1e-5)
        np.testing.assert_allclose(VT_KING_TEST_ARRAY, vt, rtol=1e-5)
        np.testing.assert_allclose(R_KING_TEST_ARRAY, r, rtol=1e-5)

    def test_set_reff(self):
        reff = CMCSAMPLECLASS.set_reff(mass=np.array([10.0, 20.0]), metallicity=0.02)
        np.testing.assert_allclose(REFF_TEST_ARRAY, reff)

    def test_cmc_sampler(self):
        np.random.seed(2)
        # Test generating CMC initial conditions and test saving the output to files
        Singles, Binaries = InitialCMCTable.sampler('cmc', binfrac_model=0.2, primary_model='kroupa01', ecc_model='sana12', porb_model='sana12', cluster_profile='plummer', met=0.014, size=20, params=os.path.join(TEST_DATA_DIR,'Params.ini'), gamma=4, r_max=100, qmin=0.1)
        InitialCMCTable.write(Singles, Binaries, filename="input.hdf5")
        InitialCMCTable.write(Singles, Binaries, filename="input.fits")
        Singles, Binaries = InitialCMCTable.read(filename="input.fits")
        # read the test files and compare to the static unit tests files
        pd.testing.assert_frame_equal(Singles, SINGLES_CMC_FITS)
        pd.testing.assert_frame_equal(Binaries, BINARIES_CMC_FITS)
        Singles, Binaries = InitialCMCTable.read(filename="input.hdf5")
        pd.testing.assert_frame_equal(Singles, SINGLES_CMC_HDF5)
        pd.testing.assert_frame_equal(Binaries, BINARIES_CMC_HDF5)



# -*- coding: utf-8 -*-
# Copyright (C) Scott Coughlin (2017 - 2021)
#
# This file is part of cosmic.
#
# cosmic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cosmic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cosmic.  If not, see <http://www.gnu.org/licenses/>.

import numpy
from . import zdata

__author__ = "Scott Coughlin <scott.coughlin@ligo.org>"
__all__ = ["zcnsts"]

c = numpy.array([3.040581e-01, 8.049509e-02, 8.967485e-02, 8.780198e-02, 2.219170e-02])


def zcnsts(z):
    """Calculate constants based on the metallicty
    *      zpars:  1; M below which hook doesn't appear on MS, Mhook.
    *              2; M above which He ignition occurs non-degenerately, Mhef.
    *              3; M above which He ignition occurs on the HG, Mfgb.
    *              4; M below which C/O ignition doesn't occur, Mup.
    *              5; M above which C ignites in the centre, Mec.
    *              6; value of log D for M<= zpars[2]
    *              7; value of x for Rgb propto M^(-x)
    *              8; value of x for tMS = numpy.maximum(tHOOK,x*tBGB)
    *              9; constant for McHeIf when computing Mc,BGB, mchefl.
    *             10; constant for McHeIf when computing Mc,HeI, mchefl.
    *             11; hydrogen abundance.
    *             12; helium abundance.
    *             13; constant x in rnumpy.minimum = rgb*x**y used by LM CHeB.
    *             14; z**0.4 to be used for WD L formula.
    """

    # initialize arrays
    msp = numpy.zeros(shape=(200, len(z)))
    zpars = numpy.zeros(shape=(20, len(z)))

    lzs = numpy.log10(z / 0.020)
    dlzs = 1.0 / (z * numpy.log(10.0))
    lz = numpy.log10(z)

    zpars[0] = 1.01850 + lzs * (0.16015 + lzs * 0.0892)
    zpars[1] = 1.9950 + lzs * (0.25 + lzs * 0.087)
    zpars[2] = 16.50 * z ** 0.06 / (1.0 + (1.0e-04 / z) ** 1.27)
    zpars[3] = numpy.maximum(6.110440 + 1.02167 * lzs, 5.0)
    zpars[4] = zpars[3] + 1.80
    zpars[5] = 5.370 + lzs * 0.135
    zpars[6] = c[0] + lzs * (c[1] + lzs * (c[2] + lzs * (c[3] + lzs * c[4])))
    zpars[7] = numpy.maximum(
        0.950,
        numpy.maximum(
            0.95 - (10.0 / 3.0) * (z - 0.01),
            numpy.minimum(0.990, 0.98 - (100.0 / 7.0) * (z - 0.001)),
        ),
    )

    # Lzams

    msp[0] = zdata.xz[0] + lzs * (
        zdata.xz[1] + lzs * (zdata.xz[2] + lzs * (zdata.xz[3] + lzs * zdata.xz[4]))
    )
    msp[1] = zdata.xz[5] + lzs * (
        zdata.xz[6] + lzs * (zdata.xz[7] + lzs * (zdata.xz[8] + lzs * zdata.xz[9]))
    )
    msp[2] = zdata.xz[10] + lzs * (
        zdata.xz[11] + lzs * (zdata.xz[12] + lzs * (zdata.xz[13] + lzs * zdata.xz[14]))
    )
    msp[3] = zdata.xz[15] + lzs * (
        zdata.xz[16] + lzs * (zdata.xz[17] + lzs * (zdata.xz[18] + lzs * zdata.xz[19]))
    )
    msp[4] = zdata.xz[20] + lzs * (
        zdata.xz[21] + lzs * (zdata.xz[22] + lzs * (zdata.xz[23] + lzs * zdata.xz[24]))
    )
    msp[5] = zdata.xz[25] + lzs * (
        zdata.xz[26] + lzs * (zdata.xz[27] + lzs * (zdata.xz[28] + lzs * zdata.xz[29]))
    )
    msp[6] = zdata.xz[30] + lzs * (
        zdata.xz[31] + lzs * (zdata.xz[32] + lzs * (zdata.xz[33] + lzs * zdata.xz[34]))
    )

    # Rzams

    msp[7] = zdata.xz[35] + lzs * (
        zdata.xz[36] + lzs * (zdata.xz[37] + lzs * (zdata.xz[38] + lzs * zdata.xz[39]))
    )
    msp[8] = zdata.xz[40] + lzs * (
        zdata.xz[41] + lzs * (zdata.xz[42] + lzs * (zdata.xz[43] + lzs * zdata.xz[44]))
    )
    msp[9] = zdata.xz[45] + lzs * (
        zdata.xz[46] + lzs * (zdata.xz[47] + lzs * (zdata.xz[48] + lzs * zdata.xz[49]))
    )
    msp[10] = zdata.xz[50] + lzs * (
        zdata.xz[51] + lzs * (zdata.xz[52] + lzs * (zdata.xz[53] + lzs * zdata.xz[54]))
    )
    msp[11] = zdata.xz[55] + lzs * (
        zdata.xz[56] + lzs * (zdata.xz[57] + lzs * (zdata.xz[58] + lzs * zdata.xz[59]))
    )
    msp[12] = zdata.xz[60]
    msp[13] = zdata.xz[61] + lzs * (
        zdata.xz[62] + lzs * (zdata.xz[63] + lzs * (zdata.xz[64] + lzs * zdata.xz[65]))
    )
    msp[14] = zdata.xz[66] + lzs * (
        zdata.xz[67] + lzs * (zdata.xz[68] + lzs * (zdata.xz[69] + lzs * zdata.xz[70]))
    )
    msp[15] = zdata.xz[71] + lzs * (
        zdata.xz[72] + lzs * (zdata.xz[73] + lzs * (zdata.xz[74] + lzs * zdata.xz[75]))
    )

    # Tbgb

    msp[16] = zdata.xt[0] + lzs * (
        zdata.xt[1] + lzs * (zdata.xt[2] + lzs * zdata.xt[3])
    )
    msp[17] = zdata.xt[4] + lzs * (
        zdata.xt[5] + lzs * (zdata.xt[6] + lzs * zdata.xt[7])
    )
    msp[18] = zdata.xt[8] + lzs * (
        zdata.xt[9] + lzs * (zdata.xt[10] + lzs * zdata.xt[11])
    )
    msp[19] = zdata.xt[12] + lzs * (
        zdata.xt[13] + lzs * (zdata.xt[14] + lzs * zdata.xt[15])
    )
    msp[20] = zdata.xt[16]

    # dTbgb/dz
    msp[116] = dlzs * (
        zdata.xt[1] + lzs * (2.0 * zdata.xt[2] + 3.0 * lzs * zdata.xt[3])
    )
    msp[117] = dlzs * (
        zdata.xt[5] + lzs * (2.0 * zdata.xt[6] + 3.0 * lzs * zdata.xt[7])
    )
    msp[118] = dlzs * (
        zdata.xt[9] + lzs * (2.0 * zdata.xt[10] + 3.0 * lzs * zdata.xt[11])
    )
    msp[119] = dlzs * (
        zdata.xt[13] + lzs * (2.0 * zdata.xt[14] + 3.0 * lzs * zdata.xt[15])
    )

    # Thook
    msp[21] = zdata.xt[17] + lzs * (
        zdata.xt[18] + lzs * (zdata.xt[19] + lzs * zdata.xt[20])
    )
    msp[22] = zdata.xt[21]
    msp[23] = zdata.xt[22] + lzs * (
        zdata.xt[23] + lzs * (zdata.xt[24] + lzs * zdata.xt[25])
    )
    msp[24] = zdata.xt[26] + lzs * (
        zdata.xt[27] + lzs * (zdata.xt[28] + lzs * zdata.xt[29])
    )
    msp[25] = zdata.xt[30]

    # Ltms
    msp[26] = zdata.xl[0] + lzs * (
        zdata.xl[1] + lzs * (zdata.xl[2] + lzs * (zdata.xl[3] + lzs * zdata.xl[4]))
    )
    msp[27] = zdata.xl[5] + lzs * (
        zdata.xl[6] + lzs * (zdata.xl[7] + lzs * (zdata.xl[8] + lzs * zdata.xl[9]))
    )
    msp[28] = zdata.xl[10] + lzs * (
        zdata.xl[11] + lzs * (zdata.xl[12] + lzs * zdata.xl[13])
    )
    msp[29] = zdata.xl[14] + lzs * (
        zdata.xl[15] + lzs * (zdata.xl[16] + lzs * (zdata.xl[17] + lzs * zdata.xl[18]))
    )
    msp[26] = msp[26] * msp[29]
    msp[27] = msp[27] * msp[29]
    msp[30] = zdata.xl[19] + lzs * (
        zdata.xl[20] + lzs * (zdata.xl[21] + lzs * zdata.xl[22])
    )
    msp[31] = zdata.xl[23] + lzs * (
        zdata.xl[24] + lzs * (zdata.xl[25] + lzs * zdata.xl[26])
    )

    # Lalpha
    m2 = 2.0
    msp[32] = zdata.xl[27] + lzs * (
        zdata.xl[28] + lzs * (zdata.xl[29] + lzs * zdata.xl[30])
    )
    msp[33] = zdata.xl[31] + lzs * (
        zdata.xl[32] + lzs * (zdata.xl[33] + lzs * zdata.xl[34])
    )
    msp[34] = zdata.xl[35] + lzs * (
        zdata.xl[36] + lzs * (zdata.xl[37] + lzs * zdata.xl[38])
    )
    msp[35] = zdata.xl[39] + lzs * (
        zdata.xl[40] + lzs * (zdata.xl[41] + lzs * zdata.xl[42])
    )
    msp[36] = numpy.maximum(0.90, 1.1064 + lzs * (0.415 + 0.18 * lzs))
    msp[37] = numpy.maximum(1.0, 1.19 + lzs * (0.377 + 0.176 * lzs))

    if (z > 0.010).any():
        msp[36, z > 0.010] = numpy.minimum(msp[36, z > 0.010], 1.0)
        msp[37, z > 0.010] = numpy.minimum(msp[37, z > 0.010], 1.10)

    msp[38] = numpy.maximum(0.1450, 0.0977 - lzs * (0.231 + 0.0753 * lzs))
    msp[39] = numpy.minimum(0.240 + lzs * (0.18 + 0.595 * lzs), 0.306 + 0.053 * lzs)
    msp[40] = numpy.minimum(0.330 + lzs * (0.132 + 0.218 * lzs), 0.36250 + 0.062 * lzs)
    msp[41] = (msp[32] + msp[33] * m2 ** msp[35]) / (m2 ** 0.40 + msp[34] * m2 ** 1.9)

    # Lbeta
    msp[42] = zdata.xl[43] + lzs * (
        zdata.xl[44] + lzs * (zdata.xl[45] + lzs * (zdata.xl[46] + lzs * zdata.xl[47]))
    )
    msp[43] = zdata.xl[48] + lzs * (
        zdata.xl[49] + lzs * (zdata.xl[50] + lzs * (zdata.xl[51] + lzs * zdata.xl[52]))
    )
    msp[44] = zdata.xl[53] + lzs * (zdata.xl[54] + lzs * zdata.xl[55])
    msp[45] = numpy.minimum(1.40, 1.5135 + 0.3769 * lzs)
    msp[45] = numpy.maximum(0.63550 - 0.4192 * lzs, numpy.maximum(1.25, msp[45]))

    # Lhook
    msp[46] = zdata.xl[56] + lzs * (
        zdata.xl[57] + lzs * (zdata.xl[58] + lzs * zdata.xl[59])
    )
    msp[47] = zdata.xl[60] + lzs * (
        zdata.xl[61] + lzs * (zdata.xl[62] + lzs * zdata.xl[63])
    )
    msp[48] = zdata.xl[64] + lzs * (
        zdata.xl[65] + lzs * (zdata.xl[66] + lzs * zdata.xl[67])
    )
    msp[49] = zdata.xl[68] + lzs * (
        zdata.xl[69] + lzs * (zdata.xl[70] + lzs * zdata.xl[71])
    )
    msp[50] = numpy.minimum(1.40, 1.5135 + 0.3769 * lzs)
    msp[50] = numpy.maximum(0.63550 - 0.4192 * lzs, numpy.maximum(1.25, msp[50]))

    # Rtms
    msp[51] = zdata.xr[0] + lzs * (
        zdata.xr[1] + lzs * (zdata.xr[2] + lzs * (zdata.xr[3] + lzs * zdata.xr[4]))
    )
    msp[52] = zdata.xr[5] + lzs * (
        zdata.xr[6] + lzs * (zdata.xr[7] + lzs * (zdata.xr[8] + lzs * zdata.xr[9]))
    )
    msp[53] = zdata.xr[10] + lzs * (
        zdata.xr[11] + lzs * (zdata.xr[12] + lzs * (zdata.xr[13] + lzs * zdata.xr[14]))
    )
    msp[54] = zdata.xr[15] + lzs * (
        zdata.xr[16] + lzs * (zdata.xr[17] + lzs * zdata.xr[18])
    )
    msp[55] = zdata.xr[19] + lzs * (
        zdata.xr[20] + lzs * (zdata.xr[21] + lzs * zdata.xr[22])
    )
    msp[51] = msp[51] * msp[53]
    msp[52] = msp[52] * msp[53]
    msp[56] = zdata.xr[23]
    msp[57] = zdata.xr[24] + lzs * (
        zdata.xr[25] + lzs * (zdata.xr[26] + lzs * zdata.xr[27])
    )
    msp[58] = zdata.xr[28] + lzs * (
        zdata.xr[29] + lzs * (zdata.xr[30] + lzs * zdata.xr[31])
    )
    msp[59] = zdata.xr[32] + lzs * (
        zdata.xr[33] + lzs * (zdata.xr[34] + lzs * zdata.xr[35])
    )
    msp[60] = zdata.xr[36] + lzs * (
        zdata.xr[37] + lzs * (zdata.xr[38] + lzs * zdata.xr[39])
    )
    #
    msp[61] = numpy.maximum(
        0.0970 - 0.1072 * (lz + 3.0),
        numpy.maximum(0.097, numpy.minimum(0.1461, 0.14610 + 0.1237 * (lz + 2.0))),
    )
    msp[61] = 10.0 ** msp[61]
    m2 = msp[61] + 0.10
    msp[62] = (msp[51] + msp[52] * msp[61] ** msp[54]) / (msp[53] + msp[61] ** msp[55])
    msp[63] = (
        msp[56] * m2 ** 3 + msp[57] * m2 ** msp[60] + msp[58] * m2 ** (msp[60] + 1.50)
    ) / (msp[59] + m2 ** 5)

    # Ralpha
    msp[64] = zdata.xr[40] + lzs * (
        zdata.xr[41] + lzs * (zdata.xr[42] + lzs * zdata.xr[43])
    )
    msp[65] = zdata.xr[44] + lzs * (
        zdata.xr[45] + lzs * (zdata.xr[46] + lzs * zdata.xr[47])
    )
    msp[66] = zdata.xr[48] + lzs * (
        zdata.xr[49] + lzs * (zdata.xr[50] + lzs * zdata.xr[51])
    )
    msp[67] = zdata.xr[52] + lzs * (
        zdata.xr[53] + lzs * (zdata.xr[54] + lzs * zdata.xr[55])
    )
    msp[68] = zdata.xr[56] + lzs * (
        zdata.xr[57] + lzs * (zdata.xr[58] + lzs * (zdata.xr[59] + lzs * zdata.xr[60]))
    )
    msp[69] = numpy.maximum(0.90, numpy.minimum(1.0, 1.116 + 0.166 * lzs))
    msp[70] = numpy.maximum(
        1.4770 + 0.296 * lzs, numpy.minimum(1.6, -0.308 - 1.046 * lzs)
    )
    msp[70] = numpy.maximum(0.80, numpy.minimum(0.8 - 2.0 * lzs, msp[70]))
    msp[71] = zdata.xr[61] + lzs * (zdata.xr[62] + lzs * zdata.xr[63])
    msp[72] = numpy.maximum(0.0650, 0.0843 - lzs * (0.0475 + 0.0352 * lzs))
    msp[73] = 0.07360 + lzs * (0.0749 + 0.04426 * lzs)
    if (z < 0.0040).any():
        msp[73, z < 0.0040] = numpy.minimum(0.055, msp[73, z < 0.0040])

    msp[74] = numpy.maximum(0.0910, numpy.minimum(0.121, 0.136 + 0.0352 * lzs))
    msp[75] = (msp[64] * msp[70] ** msp[66]) / (msp[65] + msp[70] ** msp[67])

    if (msp[69] > msp[70]).any():
        msp[69, msp[69] > msp[70]] = msp[70, msp[69] > msp[70]]
        msp[74, msp[69] > msp[70]] = msp[75, msp[69] > msp[70]]

    # Rbeta
    msp[76] = zdata.xr[64] + lzs * (
        zdata.xr[65] + lzs * (zdata.xr[66] + lzs * zdata.xr[67])
    )
    msp[77] = zdata.xr[68] + lzs * (
        zdata.xr[69] + lzs * (zdata.xr[70] + lzs * zdata.xr[71])
    )
    msp[78] = zdata.xr[72] + lzs * (
        zdata.xr[73] + lzs * (zdata.xr[74] + lzs * zdata.xr[75])
    )
    msp[79] = zdata.xr[76] + lzs * (
        zdata.xr[77] + lzs * (zdata.xr[78] + lzs * zdata.xr[79])
    )
    msp[80] = zdata.xr[80] + lzs * (zdata.xr[81] + lzs * lzs * zdata.xr[82])
    if (z > 0.010).any():
        msp[80, z > 0.010] = numpy.maximum(msp[80, z > 0.010], 0.95)

    msp[81] = numpy.maximum(
        1.40, numpy.minimum(1.6, 1.6 + lzs * (0.764 + 0.3322 * lzs))
    )

    # Rgamma
    msp[82] = numpy.maximum(
        zdata.xr[83] + lzs * (zdata.xr[84] + lzs * (zdata.xr[85] + lzs * zdata.xr[86])),
        zdata.xr[95] + lzs * (zdata.xr[96] + lzs * zdata.xr[97]),
    )
    msp[83] = numpy.minimum(
        0.0,
        zdata.xr[87] + lzs * (zdata.xr[88] + lzs * (zdata.xr[89] + lzs * zdata.xr[90])),
    )
    msp[83] = numpy.maximum(
        msp[83], zdata.xr[98] + lzs * (zdata.xr[99] + lzs * zdata.xr[100])
    )
    msp[84] = zdata.xr[91] + lzs * (
        zdata.xr[92] + lzs * (zdata.xr[93] + lzs * zdata.xr[94])
    )
    msp[84] = numpy.maximum(0.0, numpy.minimum(msp[84], 7.454 + 9.046 * lzs))
    msp[85] = numpy.minimum(
        zdata.xr[101] + lzs * zdata.xr[102], numpy.maximum(2.0, -13.3 - 18.6 * lzs)
    )
    msp[86] = numpy.minimum(1.50, numpy.maximum(0.4, 2.493 + 1.1475 * lzs))
    msp[87] = numpy.maximum(1.0, numpy.minimum(1.27, 0.8109 - 0.6282 * lzs))
    msp[87] = numpy.maximum(msp[87], 0.63550 - 0.4192 * lzs)
    msp[88] = numpy.maximum(5.855420e-02, -0.27110 - lzs * (0.5756 + 0.0838 * lzs))

    # Rhook
    msp[89] = zdata.xr[103] + lzs * (
        zdata.xr[104] + lzs * (zdata.xr[105] + lzs * zdata.xr[106])
    )
    msp[90] = zdata.xr[107] + lzs * (
        zdata.xr[108] + lzs * (zdata.xr[109] + lzs * zdata.xr[110])
    )
    msp[91] = zdata.xr[111] + lzs * (
        zdata.xr[112] + lzs * (zdata.xr[113] + lzs * zdata.xr[114])
    )
    msp[92] = zdata.xr[115] + lzs * (
        zdata.xr[116] + lzs * (zdata.xr[117] + lzs * zdata.xr[118])
    )
    msp[93] = numpy.minimum(
        1.250, numpy.maximum(1.10, 1.9848 + lzs * (1.1386 + 0.3564 * lzs))
    )
    msp[94] = 0.0630 + lzs * (0.0481 + 0.00984 * lzs)
    msp[95] = numpy.minimum(1.30, numpy.maximum(0.45, 1.2 + 2.45 * lzs))

    # Lneta
    msp[96, z > 0.00090] = 10.0
    msp[96, z < 0.00090] = 20.0

    # converting the below from fortran to python still needs work
    """
    # Lbgb
    gbp[0] = zdata.xg[0]+lzs*(zdata.xg[1]+lzs*(zdata.xg[2]+lzs*zdata.xg[3]))
    gbp[1] = zdata.xg[4]+lzs*(zdata.xg[5]+lzs*(zdata.xg[6]+lzs*zdata.xg[7]))
    gbp[2] = zdata.xg[8]+lzs*(zdata.xg[9]+lzs*(zdata.xg[10]+lzs*zdata.xg[11]))
    gbp[3] = zdata.xg[12]+lzs*(zdata.xg[13]+lzs*(zdata.xg[14]+lzs*zdata.xg[15]))
    gbp[4] = zdata.xg[16]+lzs*(zdata.xg[17]+lzs*zdata.xg[18])
    gbp[5] = zdata.xg[19]+lzs*(zdata.xg[20]+lzs*zdata.xg[21])
    gbp[2] = gbp[2]**gbp[5]
    gbp[6] = zdata.xg[22]
    gbp[7] = zdata.xg[23]

    # Lbagb
    # set gbp[15] = 1.0 until it is reset later with an initial
    # call to Lbagbf using mass = zpars[1] and mhefl = 0.0
    gbp[8] = zdata.xg[24] + lzs*(zdata.xg[25] + lzs*zdata.xg[26])
    gbp[9] = zdata.xg[27] + lzs*(zdata.xg[28] + lzs*zdata.xg[29])
    gbp[10] = 15.0
    gbp[11] = zdata.xg[30]+lzs*(zdata.xg[31]+lzs*(zdata.xg[32]+lzs*zdata.xg[33]))
    gbp[12] = zdata.xg[34]+lzs*(zdata.xg[35]+lzs*(zdata.xg[36]+lzs*zdata.xg[37]))
    gbp[13] = zdata.xg[38]+lzs*(zdata.xg[39]+lzs*(zdata.xg[40]+lzs*zdata.xg[41]))
    gbp[14] = zdata.xg[42]+lzs*zdata.xg[43]
    gbp[11] = gbp[11]**gbp[14]
    gbp[13] = gbp[13]**gbp[14]
    gbp[15] = 1.0

    # Rgb
    gbp[16] = -4.67390-0.9394*lz
    gbp[16] = 10.0**gbp[16]
    gbp[16] = numpy.maximum(gbp[16],-0.041670+55.67*z)
    gbp[16] = numpy.minimum(gbp[16],0.47710-9329.21*z**2.94)
    gbp[17] = numpy.minimum(0.540,0.397+lzs*(0.28826+0.5293*lzs))
    gbp[18] = numpy.maximum(-0.14510,-2.2794-lz*(1.5175+0.254*lz))
    gbp[18] = 10.0**gbp[18]
    if (z > 0.0040):
        gbp[18] = numpy.maximum(gbp[18],0.73070+14265.1*z**3.395)

    gbp[19] = zdata.xg[44]+lzs*(zdata.xg[45]+lzs*(zdata.xg[46]+lzs*(zdata.xg[47]+lzs*(zdata.xg[48]+lzs*zdata.xg[49]))))
    gbp[20] = zdata.xg[50]+lzs*(zdata.xg[51]+lzs*(zdata.xg[52]+lzs*(zdata.xg[53]+lzs*zdata.xg[54])))
    gbp[21] = zdata.xg[55]+lzs*(zdata.xg[56]+lzs*(zdata.xg[57]+lzs*(zdata.xg[58]+lzs*(zdata.xg[59]+lzs*zdata.xg[60]))))
    gbp[22] = zdata.xg[61]+lzs*(zdata.xg[62]+lzs*(zdata.xg[63]+lzs*(zdata.xg[64]+lzs*zdata.xg[65])))

    # Ragb
    gbp[23] = numpy.minimum(0.991640-743.123*z**2.83,
                  1.04220+lzs*(0.13156+0.045*lzs))
    gbp[24] = zdata.xg[66]+lzs*(zdata.xg[67]+lzs*(zdata.xg[68]+lzs*(zdata.xg[69]+ lzs*(zdata.xg[70]+lzs*zdata.xg[71]))))
    gbp[25] = zdata.xg[72]+lzs*(zdata.xg[73]+lzs*(zdata.xg[74]+lzs*(zdata.xg[75]+lzs*zdata.xg[76])))
    gbp[26] = zdata.xg[77]+lzs*(zdata.xg[78]+lzs*(zdata.xg[79]+lzs*(zdata.xg[80]+lzs*(zdata.xg[81]+lzs*zdata.xg[82]))))
    gbp[27] = zdata.xg[83]+lzs*(zdata.xg[84]+lzs*(zdata.xg[85]+lzs*(zdata.xg[86]+lzs*zdata.xg[87])))
    gbp[28] = zdata.xg[88]+lzs*(zdata.xg[89]+lzs*(zdata.xg[90]+lzs*(zdata.xg[91]+lzs*(zdata.xg[92]+lzs*zdata.xg[93]))))
    gbp[29] = zdata.xg[94]+lzs*(zdata.xg[95]+lzs*(zdata.xg[96]+lzs*(zdata.xg[97]+lzs*(zdata.xg[98]+lzs*zdata.xg[99]))))
    m1 = zpars[1] - 0.20
    gbp[30] = gbp[28] + gbp[29]*m1
    gbp[31] = numpy.minimum(gbp[24]/zpars[1]**gbp[25],gbp[26]/zpars[1]**gbp[27])

    # Mchei
    gbp[32] = zdata.xg[100]**4
    gbp[33] = zdata.xg[101]*4.0

    # Mcagb
    gbp[34] = zdata.xg[102]+lzs*(zdata.xg[103]+lzs*(zdata.xg[104]+lzs*zdata.xg[105]))
    gbp[35] = zdata.xg[106]+lzs*(zdata.xg[107]+lzs*(zdata.xg[108]+lzs*zdata.xg[109]))
    gbp[36] = zdata.xg[110]+lzs*zdata.xg[111]
    gbp[34] = gbp[34]**4
    gbp[35] = gbp[35]*4.0
    gbp[36] = gbp[36]**4

    # Lhei
    # set gbp[40] = -1.0 until it is reset later with an initial
    # call to Lheif using mass = zpars[1] and mhefl = 0.0
    gbp[37] = zdata.xh[0]+lzs*zdata.xh[1]
    gbp[38] = zdata.xh[2]+lzs*zdata.xh[3]
    gbp[39] = zdata.xh[4]
    gbp[40] = -1.0
    gbp[41] = zdata.xh[5]+lzs*(zdata.xh[6]+lzs*zdata.xh[7])
    gbp[42] = zdata.xh[8]+lzs*(zdata.xh[9]+lzs*zdata.xh[10])
    gbp[43] = zdata.xh[11]+lzs*(zdata.xh[12]+lzs*zdata.xh[13])
    gbp[41] = gbp[41]**2
    gbp[43] = gbp[43]**2
    # Lhe
    gbp[44] = zdata.xh[14]+lzs*(zdata.xh[15]+lzs*zdata.xh[16])
    if (lzs > -1.0):
        gbp[45] = 1.0 - zdata.xh[18]*(lzs+1.0)**zdata.xh[17]
    else:
        gbp[45] = 1.0

    gbp[46] = zdata.xh[19]+lzs*(zdata.xh[20]+lzs*zdata.xh[21])
    gbp[47] = zdata.xh[22]+lzs*(zdata.xh[23]+lzs*zdata.xh[24])
    gbp[44] = gbp[44]**gbp[47]
    gbp[46] = gbp[46]**gbp[47]
    gbp[45] = gbp[45]/zpars[2]**0.10+(gbp[45]*gbp[46]-gbp[44])/zpars[2]**(gbp[47]+0.10)

    # Rnumpy.minimum

    gbp[48] = zdata.xh[25]+lzs*(zdata.xh[26]+lzs*(zdata.xh[27]+lzs*zdata.xh[28]))
    gbp[49] = zdata.xh[29]+lzs*(zdata.xh[30]+lzs*(zdata.xh[31]+lzs*zdata.xh[32]))
    gbp[50] = zdata.xh[33]+lzs*(zdata.xh[34]+lzs*(zdata.xh[35]+lzs*zdata.xh[36]))
    gbp[51] = 5.0+zdata.xh[37]*z**zdata.xh[38]
    gbp[52] = zdata.xh[39]+lzs*(zdata.xh[40]+lzs*(zdata.xh[41]+lzs*zdata.xh[42]))
    gbp[48] = gbp[48]**gbp[52]
    gbp[50] = gbp[50]**(2.0*gbp[52])

    # The
    # set gbp[56] = -1.0 until it is reset later with an initial
    # call to Thef using mass = zpars[1], mc = 0.0  and mhefl = 0.0
    gbp[53] = zdata.xh[43]+lzs*(zdata.xh[44]+lzs*(zdata.xh[45]+lzs*zdata.xh[46]))
    gbp[54] = zdata.xh[47]+lzs*(zdata.xh[48]+lzs*zdata.xh[49])
    gbp[54] = numpy.maximum(gbp[54],1.0)
    gbp[55] = zdata.xh[50]
    gbp[56] = -1.0
    gbp[57] = zdata.xh[51]+lzs*(zdata.xh[52]+lzs*(zdata.xh[53]+lzs*zdata.xh[54]))
    gbp[58] = zdata.xh[55]+lzs*(zdata.xh[56]+lzs*(zdata.xh[57]+lzs*zdata.xh[58]))
    gbp[59] = zdata.xh[59]+lzs*(zdata.xh[60]+lzs*(zdata.xh[61]+lzs*zdata.xh[62]))
    gbp[60] = zdata.xh[63]+lzs*zdata.xh[64]
    gbp[57] = gbp[57]**gbp[60]
    gbp[59] = gbp[59]**5

    # Tbl
    dum1 = zpars[1]/zpars[2]
    gbp[61] = zdata.xh[65]+lzs*zdata.xh[66]
    gbp[61] = -gbp[61]*numpy.log10(dum1)
    gbp[62] = zdata.xh[67]
    if (lzd > 0.0):
        gbp[63] = 1.0-lzd*(zdata.xh[68]+lzd*(zdata.xh[69]+lzd*zdata.xh[70]))
    else:
        gbp[63] = 1.0

    gbp[64] = 1.0-gbp[63]*dum1**gbp[62]
    gbp[65] = 1.0 - lzd*(zdata.xh[76] + lzd*(zdata.xh[77] + lzd*zdata.xh[78]))
    gbp[66] = zdata.xh[71] + lzs*(zdata.xh[72] + lzs*(zdata.xh[73] + lzs*zdata.xh[74]))
    gbp[67] = zdata.xh[75]

    # Lzahb
    gbp[68] = zdata.xh[79] + lzs*(zdata.xh[80] + lzs*zdata.xh[81])
    gbp[69] = zdata.xh[82] + lzs*(zdata.xh[83] + lzs*zdata.xh[84])
    gbp[70] = 15.0
    gbp[71] = zdata.xh[85]
    gbp[72] = zdata.xh[86]

    # Rzahb
    gbp[74] = zdata.xh[87] + lzs*(zdata.xh[88] + lzs*(zdata.xh[89] + lzs*zdata.xh[90]))
    gbp[75] = zdata.xh[91] + lzs*(zdata.xh[92] + lzs*(zdata.xh[93] + lzs*zdata.xh[94]))
    gbp[76] = zdata.xh[95] + lzs*(zdata.xh[96] + lzs*(zdata.xh[97] + lzs*zdata.xh[98]))

    # finish Lbagb
    mhefl = 0.0
    lx = lbagbf(zpars[1], mhefl)
    gbp[15] = lx

    # finish LHeI
    dum1 = 0.0
    lhefl = lheif(zpars[1],mhefl)
    gbp[40] = (gbp[37]*zpars[1]**gbp[38]-lhefl)/(numpy.exp(zpars[1]*gbp[39])*lhefl)

    # finish THe
    thefl = thef(zpars[1],dum1,mhefl)*tbgbf(zpars[1])
    gbp[56] = (thefl-gbp[53])/(gbp[53]*numpy.exp(gbp[55]*zpars[1]))

    # finish Tblf
    rb = ragbf(zpars[2],lheif(zpars[2],zpars[1]),mhefl)
    rr = 1.0 - rnumpy.minimumf(zpars[2])/rb
    rr = numpy.maximum(rr,1.0e-12)
    gbp[65] = gbp[65]/(zpars[2]**gbp[66]*rr**gbp[67])

    # finish Lzahb
    gbp[73] = lhefl*lHef(zpars[1])

    kw = 0
    tm = 0.0
    tn = 0.0
    star(kw,zpars[1],zpars[1],tm,tn,tscls,lums,GB,zpars)
    zpars[8] = mcgbf(lums[2],GB,lums[5])
    zpars[9] = mcgbf(lums[3],GB,lums[5])

    # set the hydrogen and helium abundances
    zpars[10] = 0.760 - 3.0*z
    zpars[11] = 0.240 + 2.0*z

    # set constant for low-mass CHeB stars
    zpars[12] = (rnumpy.minimumf(zpars[1])/
                rgbf(zpars[1],lzahbf(zpars[1],zpars[8],zpars[1])))

    zpars[13] = z**0.40
    #
    """
    return zpars, msp



# -*- coding: utf-8 -*-
# Copyright (C) Scott Coughlin (2017 - 2021)
#
# This file is part of cosmic.
#
# cosmic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cosmic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cosmic.  If not, see <http://www.gnu.org/licenses/>.

"""`InitialCMCTable`
"""

import pandas as pd
import numpy as np
import h5py
from astropy.io import fits
from astropy.table import Table

__author__ = "Scott Coughlin <scottcoughlin2014@u.northwestern.edu>"
__credits__ = "Carl Rodriguez <carllouisrodriguez@gmail.com>"
__all__ = ["InitialCMCTable"]


INITIAL_CONDITIONS_COLUMNS_CMC_SINGLES = [
    "id",
    "k",
    "m",
    "Reff",
    "r",
    "vr",
    "vt",
    "binind",
]

INITIAL_CONDITIONS_COLUMNS_CMC_BINARIES = [
    "index",
    "id1",
    "k1",
    "m1",
    "Reff1",
    "id2",
    "k2",
    "m2",
    "Reff2",
    "a",
    "e",
]


class InitialCMCTable(pd.DataFrame):
    scaled_to_nbody_units = False
    metallicity = None
    mass_of_cluster = None
    virial_radius = None
    tidal_radius = None

    @classmethod
    def ScaleToNBodyUnits(cls, Singles, Binaries, virial_radius=1):
        r"""Rescale the single masses, radii, and velocities into N-body units
           i.e. \sum m = M = 1
                 Kinetic Energy   = 0.25
                 Potential Energy = -0.5

        Note that this is already done for r, vr, and vt from the profile generators.
        However, after the stellar masses are assigned we need to redo it, and 
        the stellar radii and seperations need to be converted from RSUN to code units

        Parameters
        ----------
        Singles : DataFrame
            Pandas DataFrame from the InitialCMCSingles function
        Binaries : DataFrame
            Pandas DataFrame from the InitialCMCSingles function
        virial_radius : float
            Virial radius of the cluster in parsec (default 1pc)

        Returns
        -------
        None: Pandas dataframes are modified in place
        """

        # Normalize the masses to the total cluster mass
        M_total = sum(Singles["m"])
        Singles["m"] /= M_total
        Binaries["m1"] /= M_total
        Binaries["m2"] /= M_total

        # Take the radii, and offset by one
        radius = np.array(Singles["r"])
        radius_p1 = np.append(radius[1:], [1e100])

        # Masses and velocities
        mass = np.array(Singles["m"])
        cumul_mass = np.cumsum(mass)
        vr = np.array(Singles["vr"])
        vt = np.array(Singles["vt"])

        # Then compute the total kinetic and potential energy
        # There's probably a cleaner way to do the PE (this is a one-line version
        #  of the for loop we use in CMC; vectorized and pythonic, but sloppy)
        KE = 0.5 * sum(mass * (vr ** 2 + vt ** 2))
        PE = 0.5 * sum(
            mass[::-1]
            * np.cumsum((cumul_mass * (1.0 / radius - 1.0 / radius_p1))[::-1])
        )

        # Compute the position and velocity scalings
        rfac = 2 * PE
        vfac = 1.0 / np.sqrt(4 * KE)

        # Scale the positions and velocities s.t. KE=0.25, PE=-0.5
        Singles["r"] *= rfac
        Singles["vr"] *= vfac
        Singles["vt"] *= vfac

        # Finally, scale the radii and seperations from BSE into code units
        PARSEC_PER_RSUN = 2.2546101516664447e-08 
        DistConv = PARSEC_PER_RSUN / virial_radius

        Singles["Reff"] *= DistConv
        Binaries["a"] *= DistConv
        Binaries["Reff1"] *= DistConv
        Binaries["Reff2"] *= DistConv

        Singles.scaled_to_nbody_units = True
        Binaries.scaled_to_nbody_units = True
        return

    @classmethod
    def InitialCMCSingles(cls, id_idx, k, m, Reff, r, vr, vt, binind):
        """Create A Table of CMC Singles

        Parameters
        ----------
        m1 : float
            Primary mass [Msun]
        m2 : float
            Secondary mass [Msun]
        porb : float
            Orbital period [days]
        ecc : float
            Eccentricity
        kstar1 : array
            0-14 Initial stellar type of the larger object;
            main sequence stars are 0 if m < 0.7 Msun and 1 otherwise
        kstar2 : array
            0-14 Initial stellar type of the smaller object;
            main sequence stars are 0 if m < 0.7 Msun and 1 otherwise
        metallicity : float
            Metallicity of the binaries; Z_sun = 0.02

        **kwargs

            binfrac : float
                System-specific probability of the primary star being in a binary

        Returns
        -------
        InitialBinaries : DataFrame
            Single binary initial conditions

        """
        bin_dat = cls(
            np.vstack(
                [
                    id_idx,
                    k,
                    m,
                    Reff,
                    r,
                    vr,
                    vt,
                    binind,
                ]
            ).T,
            columns=INITIAL_CONDITIONS_COLUMNS_CMC_SINGLES,
        )

        return bin_dat

    @classmethod
    def InitialCMCBinaries(cls, index, id1, k1, m1, Reff1, id2, k2, m2, Reff2, a, e):
        """Create A Table of CMC Binaries

        Parameters
        ----------
        m1 : float
            Primary mass [Msun]
        m2 : float
            Secondary mass [Msun]
        porb : float
            Orbital period [days]
        ecc : float
            Eccentricity
        kstar1 : array
            0-14 Initial stellar type of the larger object;
            main sequence stars are 0 if m < 0.7 Msun and 1 otherwise
        kstar2 : array
            0-14 Initial stellar type of the smaller object;
            main sequence stars are 0 if m < 0.7 Msun and 1 otherwise
        metallicity : float
            Metallicity of the binaries; Z_sun = 0.02

        **kwargs

            binfrac : float
                System-specific probability of the primary star being in a binary

        Returns
        -------
        InitialBinaries : DataFrame
            Single binary initial conditions

        """
        bin_dat = cls(
            np.vstack([index, id1, k1, m1, Reff1, id2, k2, m2, Reff2, a, e]).T,
            columns=INITIAL_CONDITIONS_COLUMNS_CMC_BINARIES,
        )

        return bin_dat

    @classmethod
    def write(cls, Singles, Binaries, filename="input.hdf5", **kwargs):
        """Save Singles and Binaries to HDF5 or FITS file

        Parameters
        ----------
        Singles : DataFrame
            Pandas DataFrame from the InitialCMCSingles function
        Binaries : DataFrame
            Pandas DataFrame from the InitialCMCBinaries function
        filename : (str)
            Must end in ".fits" or ".hdf5/h5"

        Optional Parameteres
        --------------------
        These are automatically set in the Singles df, but can be OVERWRITTEN here 

        virial_radius : `float`
            the initial virial radius of the cluster, in parsecs

        tidal_radius : `float`
            the initial tidal radius of the cluster, in units of the virial_radius

        metallicity : `float`
            the stellar metallicity of the cluster


        Returns
        -------
            None:

        """

        # verify parameters
        if (".hdf5" in filename) or (".h5" in filename):
            savehdf5 = True
            savefits = False
        elif ".fits" in filename:
            savefits = True
            savehdf5 = False
        else:
            raise ValueError(
                "File extension not recognized, valid file types are fits and hdf5"
            )

        virial_radius = kwargs.pop('virial_radius',Singles.virial_radius)
        tidal_radius = kwargs.pop('tidal_radius',Singles.tidal_radius)
        metallicity = kwargs.pop('metallicity',Singles.metallicity)

        # If a user has not already scaled the units of the Singles and Binaries tables,
        # and the attribute mass_of_cluster is None, then
        # we can calculate it now
        if (not Singles.scaled_to_nbody_units) and (Singles.mass_of_cluster is None):
            Singles.mass_of_cluster = np.sum(Singles["m"])
            InitialCMCTable.ScaleToNBodyUnits(
                Singles, Binaries, virial_radius=virial_radius
            )
        elif (not Singles.scaled_to_nbody_units) and (Singles.mass_of_cluster is not None):
            InitialCMCTable.ScaleToNBodyUnits(
                Singles, Binaries, virial_radius=virial_radius
            )
        elif (Singles.scaled_to_nbody_units) and (Singles.mass_of_cluster is None):
            # we cannot get the pre-scaled mass of the cluster
            raise ValueError(
                "In order to save the initial conditions correctly, "
                "you cannot feed in Singles and Binaries which have already been scaled"
            )

        if Singles.metallicity is None:
            raise ValueError(
                "The user has not supplied a metallicity for the cluster. Please set the Singles.metallicity attribute"
            )

        # Need to append special rows to the start and end of the Singles table
        singles = pd.DataFrame(
            np.zeros((1, Singles.shape[1])), index=[0], columns=Singles.columns
        )
        singles_bottom = pd.DataFrame(
            np.zeros((1, Singles.shape[1])), index=[0], columns=Singles.columns
        )
        singles = pd.concat([singles, Singles])
        singles = pd.concat([singles, singles_bottom])
        singles["r"].iloc[-1] = 1e40
        singles["r"].iloc[0] = 2.2250738585072014e-308

        # Add a special row to the end of Bianries table
        binaries = pd.DataFrame(
            np.zeros((1, Binaries.shape[1])), index=[0], columns=Binaries.columns
        )
        binaries = pd.concat([binaries, Binaries])

        if savehdf5:
            singles.to_hdf(filename, key="CLUS_OBJ_DATA", mode="w")
            binaries.to_hdf(filename, key="CLUS_BINARY_DATA")
            with h5py.File(filename, "a") as f:
                f["CLUS_OBJ_DATA/block0_values"].attrs["EXTNAME"] = "CLUS_OBJ_DATA"
                f["CLUS_OBJ_DATA/block0_values"].attrs["NOBJ"] = int(len(singles)) - 2
                f["CLUS_OBJ_DATA/block0_values"].attrs["NBINARY"] = (
                    int(len(binaries)) - 1
                )
                f["CLUS_OBJ_DATA/block0_values"].attrs[
                    "MCLUS"
                ] = Singles.mass_of_cluster
                f["CLUS_OBJ_DATA/block0_values"].attrs["RVIR"] = virial_radius 
                f["CLUS_OBJ_DATA/block0_values"].attrs["RTID"] = tidal_radius 
                f["CLUS_OBJ_DATA/block0_values"].attrs["Z"] = metallicity

        if savefits:
            Singles_fits = Table.from_pandas(singles)
            Binaries_fits = Table.from_pandas(binaries)

            hdu1 = fits.table_to_hdu(Singles_fits)
            hdu2 = fits.table_to_hdu(Binaries_fits)

            # create a header
            hdr = fits.Header()
            hdr["COMMENT"] = "CMC Configured Initial Conditions"
            hdr["COMMENT"] = "Produced by COSMIC"
            primary_hdu = fits.PrimaryHDU(header=hdr)

            hdu1.header["EXTNAME"] = "CLUS_OBJ_DATA"
            hdu1.header["NOBJ"] = int(len(Singles_fits)) - 2
            hdu1.header["NBINARY"] = int(len(Binaries_fits)) - 1
            hdu1.header["MCLUS"] = Singles.mass_of_cluster
            hdu1.header["RVIR"] = virial_radius
            hdu1.header["RTID"] = tidal_radius
            hdu1.header["Z"] = metallicity

            # put all the HDUs together
            hdul = fits.HDUList([primary_hdu, hdu1, hdu2])

            # write it out
            hdul.writeto(
                filename,
                overwrite=True,
            )

        return

    @classmethod
    def read(cls, filename):
        """Read Singles and Binaries to HDF5 or FITS file

        Parameters
        ----------
        filename : (str)
            Must end in ".fits" or ".hdf5/h5"

        Returns
        -------
        Singles : DataFrame
            Pandas DataFrame from the InitialCMCSingles function
        Binaries : DataFrame
            Pandas DataFrame from the InitialCMCBinaries function
        """
        # verify parameters
        if (".hdf5" in filename) or (".h5" in filename):
            savehdf5 = True
            savefits = False
        elif ".fits" in filename:
            savefits = True
            savehdf5 = False
        else:
            raise ValueError(
                "File extension not recognized, valid file types are fits and hdf5"
            )

        if savehdf5:
            Singles = pd.read_hdf(filename, "CLUS_OBJ_DATA")
            Binaries = pd.read_hdf(filename, "CLUS_BINARY_DATA")
        elif savefits:
            Singles = cls(Table.read(filename,hdu=1).to_pandas())
            Binaries = cls(Table.read(filename, hdu=2).to_pandas())
        return Singles, Binaries


    @classmethod
    def sampler(cls, format_, *args, **kwargs):
        """Fetch a method to generate an initial binary sample

        Parameters
        ----------
        format : str
            the method name; Choose from 'independent' or 'multidim'

        *args
            the arguments necessary for the registered sample
            method; see help(InitialCMCTable.sampler('independent')
            to see the arguments necessary for the independent sample
        """
        # standard registered fetch
        from .sampler.sampler import get_sampler

        sampler = get_sampler(format_, cls)
        return sampler(*args, **kwargs)



# -*- coding: utf-8 -*-
# Copyright (C) Scott Coughlin (2017 - 2021)
#
# This file is part of cosmic.
#
# cosmic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cosmic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cosmic.  If not, see <http://www.gnu.org/licenses/>.

"""`utils`
"""
import scipy
import numpy as np
import pandas as pd
import scipy.special as ss
import astropy.stats as astrostats
import warnings
import ast
import operator
import json
import itertools
import os.path

from configparser import ConfigParser
from .bse_utils.zcnsts import zcnsts

__author__ = "Katelyn Breivik <katie.breivik@gmail.com>"
__credits__ = [
    "Scott Coughlin <scott.coughlin@ligo.org>",
    "Michael Zevin <zevin@northwestern.edu>",
]
__all__ = [
    "filter_bin_state",
    "conv_select",
    "mass_min_max_select",
    "idl_tabulate",
    "rndm",
    "param_transform",
    "dat_transform",
    "dat_un_transform",
    "knuth_bw_selector",
    "error_check",
    "check_initial_conditions",
    "convert_kstar_evol_type",
    "parse_inifile",
    "pop_write",
    "a_from_p",
    "p_from_a",
    "get_Z_from_FeH",
    "get_FeH_from_Z",
    "get_binfrac_of_Z",
    "get_porb_norm",
    "get_met_dep_binfrac"
]


def filter_bin_state(bcm, bpp, method, kstar1_range, kstar2_range):
    """Filter the output of bpp and bcm, where the kstar ranges
    have already been selected by the conv_select module

    Parameters
    ----------
    bcm : `pandas.DataFrame`
        bcm dataframe

    bpp : `pandas.DataFrame`
        bpp dataframe

    method : `dict`,
        one or more methods by which to filter the
        bpp or bcm table, e.g. ``{'binary_state' : [0,1]}``;
        This means you do *not* want to select the final state of the binaries in the bcm array

    kstar1_range : `list`
        list containing all kstar1 values to retain

    kstar2_range : `list`
        list containing all kstar2 values to retain

    Returns
    -------
    bcm : `pandas.DataFrame`
        filtered bcm dataframe
    """
    _known_methods = ["binary_state", "timestep_conditions"]

    if not set(method.keys()).issubset(set(_known_methods)):
        raise ValueError(
            "You have supplied an "
            "unknown method to filter out "
            "the bpp or bcm array. Known methods are "
            "{0}".format(_known_methods)
        )

    for meth, use in method.items():
        if meth == "binary_state":
            bin_num_save = []

            # in order to filter on binary state we need the last entry of the bcm array for each binary
            bcm_last_entry = bcm.groupby("bin_num").last().reset_index()

            # in order to find the properities of disrupted or systems
            # that are alive today we can simply check the last entry in the bcm
            # array for the system and see what its properities are today
            bcm_0_2 = bcm_last_entry.loc[(bcm_last_entry.bin_state != 1)]
            bin_num_save.extend(bcm_0_2.bin_num.tolist())

            # in order to find the properities of merged systems
            # we actually need to search in the BPP array for the properities
            # of the objects right at merge because the bcm will report
            # the post merge object only
            bcm_1 = bcm_last_entry.loc[bcm_last_entry.bin_state == 1]

            # We now find the product of the kstar range lists so we can match the
            # merger_type column from the bcm array which tells us what objects
            # merged
            merger_objects_to_track = []
            merger_objects_to_track.extend(
                list(
                    map(
                        lambda x: "{0}{1}".format(
                            str(x[0]).zfill(2), str(x[1]).zfill(2)
                        ),
                        list(itertools.product(kstar1_range, kstar2_range)),
                    )
                )
            )
            merger_objects_to_track.extend(
                list(
                    map(
                        lambda x: "{0}{1}".format(
                            str(x[0]).zfill(2), str(x[1]).zfill(2)
                        ),
                        list(itertools.product(kstar2_range, kstar1_range)),
                    )
                )
            )
            bin_num_save.extend(
                bcm_1.loc[
                    bcm_1.merger_type.isin(merger_objects_to_track)
                ].bin_num.tolist()
            )

            bcm_last_entry = bcm_last_entry.loc[
                bcm_last_entry.bin_num.isin(bin_num_save)
            ]

            # this will tell use the binary state fraction of the systems with a certain final kstar type
            # before we throw out certain binary states if a user requested that.
            bin_state_fraction = bcm_last_entry.groupby("bin_state").tphys.count()
            bin_states = []
            for ii in range(3):
                try:
                    bin_states.append(bin_state_fraction.loc[ii])
                except Exception:
                    bin_states.append(0)
            bin_state_fraction = pd.DataFrame([bin_states], columns=[0, 1, 2])

            bcm = bcm.loc[
                bcm.bin_num.isin(
                    bcm_last_entry.loc[bcm_last_entry.bin_state.isin(use)].bin_num
                )
            ]

    return bcm, bin_state_fraction


def conv_select(bcm_save, bpp_save, final_kstar_1, final_kstar_2, method, conv_lims):
    """Select bcm data for special convergence cases

    Parameters
    ----------
    bcm_save : `pandas.DataFrame`
        bcm dataframe containing all saved bcm data

    bpp_save : `pandas.DataFrame`
        bpp dataframe containing all saved bpp data

    final_kstar_1 : `list`
        contains list of final primary kstars specified by user

    final_kstar_2 : `list`
        contains list of final primary kstars specified by user

    method : `str`
        stage in binary evolution to check convergence for
        only one method may be supplied and they are specified
        in the inifile

    conv_lims : `dict`
        dictionary where keys are convergence params and the
        values are lists containing a [lo, hi] value to filter the
        convergence param between
        any non-specified convergence params will not be filtered

    Returns
    -------
    conv_save : `pandas.DataFrame`
        filtered dataframe containing binaries that fulfill
        user-specified convergence criteria

    """
    _known_methods = [
        "formation",
        "1_SN",
        "2_SN",
        "disruption",
        "final_state",
        "XRB_form",
    ]

    if method not in _known_methods:
        raise ValueError(
            "You have supplied an "
            "unknown method to filter the "
            "bcm array for convergence. Known methods are "
            "{0}".format(_known_methods)
        )

    if method == "formation":
        # filter the bpp array to find the systems that match the user-specified
        # final kstars
        conv_save = bpp_save.loc[
            ((bpp_save.kstar_1.isin(final_kstar_1)) 
             & (bpp_save.kstar_2.isin(final_kstar_2))
            )
            |
            ((bpp_save.kstar_1.isin(final_kstar_2))
             & (bpp_save.kstar_2.isin(final_kstar_1))
            )
        ]

        # select the formation parameters
        conv_save = conv_save.groupby("bin_num").first().reset_index()

    elif method == "1_SN":
        # select out the systems which will undergo a supernova
        conv_sn_ind = bpp_save.loc[bpp_save.evol_type.isin([15.0, 16.0])].bin_num

        # select out the systems which will produce the user specified final kstars
        # and undergo a supernova
        conv_sn_ind = bpp_save.loc[
            (bpp_save.bin_num.isin(conv_sn_ind))
            & (bpp_save.kstar_1.isin(final_kstar_1))
            & (bpp_save.kstar_2.isin(final_kstar_2))
            & (bpp_save.sep > 0)
        ].bin_num

        # select out the values just before the supernova(e)
        conv_sn = bpp_save.loc[
            (bpp_save.bin_num.isin(conv_sn_ind))
            & (bpp_save.evol_type.isin([15.0, 16.0]))
        ]

        # make sure to select out only the first supernova
        conv_save = conv_sn.groupby("bin_num").first().reset_index()

    elif method == "2_SN":
        # select out the systems which will undergo a supernova
        conv_sn_ind = bpp_save.loc[bpp_save.evol_type.isin([15.0, 16.0])].bin_num

        # select out the systems which will produce the user specified final kstars
        # and undergo a supernova
        conv_sn_ind = bpp_save.loc[
            (bpp_save.bin_num.isin(conv_sn_ind))
            & (bpp_save.kstar_1.isin(final_kstar_1))
            & (bpp_save.kstar_2.isin(final_kstar_2))
            & (bpp_save.sep > 0)
        ].bin_num
        # select out the values just before the supernova(e)
        conv_sn = bpp_save.loc[
            (bpp_save.bin_num.isin(conv_sn_ind))
            & (bpp_save.evol_type.isin([15.0, 16.0]))
        ]

        # select out only the systems that go through 2 supernovae
        conv_sn_2 = conv_sn.loc[conv_sn.groupby("bin_num").size() == 2]

        # make sure to select out only the second supernova
        conv_save = conv_sn_2.groupby("bin_num").nth(1).reset_index()

    elif method == "disruption":
        # filter the bpp array to find the systems that match the user-specified
        # final kstars
        conv_ind = bpp_save.loc[
            (bpp_save.kstar_1.isin(final_kstar_1))
            & (bpp_save.kstar_2.isin(final_kstar_2))
        ].bin_num.unique()

        conv_save = bpp_save.loc[(bpp_save.bin_num.isin(conv_ind))]

        # select out the parameters just before disruption
        # first reset the index:
        conv_save_reset = conv_save.reset_index()

        # next select out the index for the disrupted systems using evol_type == 11
        conv_save_reset_ind = conv_save_reset.loc[
            conv_save_reset.evol_type == 11.0
        ].index

        conv_save = conv_save_reset.iloc[conv_save_reset_ind]

    elif method == "final_state":
        # the bcm array is all that we need!
        conv_save = bcm_save

    elif method == "XRB_form":
        # select out the systems which undergo a SN
        conv_ind = bpp_save.loc[bpp_save.evol_type.isin([15.0, 16.0])].bin_num.unique()
        conv_sn = bpp_save.loc[bpp_save.bin_num.isin(conv_ind)]

        # select out systems when they first enter RLO after the 1st SN
        conv_xrb = conv_sn.loc[
            (conv_sn.kstar_1.isin(final_kstar_1))
            & (conv_sn.kstar_2.isin(final_kstar_2))
            & (conv_sn.RRLO_2 >= 1.0)
            & (conv_sn.sep > 0)
        ]
        conv_save = conv_xrb.groupby("bin_num").first().reset_index()

    if conv_lims:
        for key in conv_lims.keys():
            filter_lo = conv_lims[key][0]
            filter_hi = conv_lims[key][1]
            conv_save_lim = conv_save.loc[conv_save[key] < filter_hi]
            conv_lims_bin_num = conv_save_lim.loc[conv_save[key] > filter_lo].bin_num
    else:
        conv_lims_bin_num = conv_save.bin_num

    return conv_save, conv_lims_bin_num


def pop_write(
    dat_store,
    log_file,
    mass_list,
    number_list,
    bcm,
    bpp,
    initC,
    conv,
    kick_info,
    bin_state_nums,
    match,
    idx,
):
    """Writes all the good stuff that you want to save from runFixedPop in a
       single function

    Parameters
    ----------
    dat_store : `pandas HDFStore`
        H5 file to write to

    log_file : `file write`
        log file to write to
    mass_list : `list`
        list containing the mass of the singles, mass of the binaries,
        and mass of the stars

    n_list : `list`
        list containing the number of singles, number of binaries,
        and number of stars

    bcm : `pandas.DataFrame`
        bcm array to write

    bpp : `pandas.DataFrame`
        bpp array to write

    initCond : `pandas.DataFrame`
        initCond array to write

    conv : `pandas.DataFrame`
        conv array to write

    kick_info : `pandas.DataFrame`
        kick_info array to write

    bin_state_nums : `list`
        contains the count of binstates 0,1,2

    match : pandas.DataFrame
        contains the match values for each conv_param

    idx : `int`
        contains the index of the bcm so we can pick up where we left off
        if runFixedPop hits a wall time

    Returns
    -------
    Nothing!
    """

    m_keys = ["mass_singles", "mass_binaries", "mass_stars"]
    n_keys = ["n_singles", "n_binaries", "n_stars"]
    for m_write, m_key, n_write, n_key in zip(mass_list, m_keys, number_list, n_keys):
        # save the total_sampled_mass so far
        dat_store.append(m_key, pd.DataFrame([m_write]))
        dat_store.append(n_key, pd.DataFrame([n_write]))
    log_file.write("The total mass sampled so far is: {0}\n".format(mass_list[2]))

    # Save the bcm dataframe
    dat_store.append("bcm", bcm)

    # Save the bpp dataframe
    dat_store.append("bpp", bpp)

    # Save the initial binaries
    # ensure that the index corresponds to bin_num
    dat_store.append("initCond", initC.set_index("bin_num", drop=False))

    # Save the converging dataframe
    dat_store.append("conv", conv)

    # Save the converging dataframe
    dat_store.append("kick_info", kick_info)

    # Save number of systems in each bin state
    dat_store.append("bin_state_nums", bin_state_nums)

    # Save the matches
    dat_store.append("match", match)

    # Save the index
    dat_store.append("idx", pd.DataFrame([idx]))
    return


def a_from_p(p, m1, m2):
    """Computes the separation from orbital period with KEPLER III

    Parameters
    ----------
    p : float/array
        orbital period [day]
    m1 : float/array
        primary mass [msun]
    m2 : float/array
        secondary mass [msun]

    Returns
    -------
    sep : float/array
        separation [rsun]
    """

    p_yr = p / 365.25
    sep_3 = p_yr ** 2 * (m1 + m2)
    sep = sep_3 ** (1 / 3.0)
    sep_rsun = sep * 215.032
    return sep_rsun


def p_from_a(sep, m1, m2):
    """Computes separation from orbital period with kepler III

    Parameters
    ----------
    sep : float/array
        separation [rsun]
    m1 : float/array
        primary mass [msun]
    m2 : float/array
        secondary mass [msun]

    Returns
    -------
    p : float/array
        orbital period [day]
    """

    sep_au = sep / 215.032
    p_2 = sep_au ** 3 / (m1 + m2)
    p_day = (p_2 ** 0.5) * 365.25
    return p_day


def calc_Roche_radius(M1, M2, A):
    """Get Roche lobe radius (Eggleton 1983)

    Parameters
    ----------
    M1 : float
        Primary mass [any unit]
    M2 : float
        Secondary mass [any unit]
    A : float
        Orbital separation [any unit]

    Returns
    -------
    Roche radius : float
        in units of input 'A'
    """
    q = M1 / M2
    return (
        A
        * 0.49
        * q ** (2.0 / 3.0)
        / (0.6 * q ** (2.0 / 3.0) + np.log(1.0 + q ** (1.0 / 3.0)))
    )


def mass_min_max_select(kstar_1, kstar_2, **kwargs):
    """Select a minimum and maximum mass to filter out binaries in the initial
    parameter sample to reduce the number of unneccessary binaries evolved
    in BSE

    Parameters
    ----------
    kstar_1 : int, list
        BSE stellar type for the primary
        or minimum and maximum stellar types for the primary
    kstar_2 : int, list
        BSE stellar type for the secondary
        or minimum and maximum stellar types for the secondary

    Returns
    -------
    min_mass[0] : float
        minimum primary mass for initial sample
    max_mass[0] : float
        maximum primary mass for initial sample
    min_mass[1] : float
        minimum secondary mass for initial sample
    max_mass[1] : float
        maximum secondary mass for initial sample
    """

    primary_max = kwargs["m_max"] if "m_max" in kwargs.keys() else 150.0
    secondary_max = kwargs["m_max"] if "m_max" in kwargs.keys() else 150.0

    primary_min = kwargs["m1_min"] if "m1_min" in kwargs.keys() else 0.08
    secondary_min = kwargs["m2_min"] if "m2_min" in kwargs.keys() else 0.08

    if ((primary_min < 0.08) | (secondary_min < 0.08)):
        warnings.warn("Tread carefully, BSE is not equipped to handle stellar masses less than 0.08 Msun!")
    if primary_max > 150:
        warnings.warn("Tread carefully, BSE is not equipped to handle stellar masses greater than 150 Msun!")

    min_mass = [primary_min, secondary_min]
    max_mass = [primary_max, secondary_max]

    if len(kstar_1) == 1:
        # there is a range of final kstar_1s to save
        kstar_1_lo = kstar_1[0]
        kstar_1_hi = kstar_1[0]
    else:
        kstar_1_lo = min(kstar_1)
        kstar_1_hi = max(kstar_1)

    if len(kstar_2) == 1:
        # there is a range of final kstar_1s to save
        kstar_2_lo = kstar_2[0]
        kstar_2_hi = kstar_2[0]
    else:
        kstar_2_lo = min(kstar_2)
        kstar_2_hi = max(kstar_2)

    kstar_lo = [kstar_1_lo, kstar_2_lo]
    kstar_hi = [kstar_1_hi, kstar_2_hi]

    ii = 0
    for k in kstar_lo:
        if k == 14.0:
            min_mass[ii] = 8.0
        elif k == 13.0:
            min_mass[ii] = 3.0
        elif k == 12.0:
            min_mass[ii] = 1.0
        elif k == 11.0:
            min_mass[ii] = 0.8
        elif k == 10.0:
            min_mass[ii] = 0.5
        ii += 1

    ii = 0
    for k in kstar_hi:
        if k == 13.0:
            max_mass[ii] = 60.0
        elif k == 12.0:
            max_mass[ii] = 20.0
        elif k == 11.0:
            max_mass[ii] = 20.0
        elif k == 10.0:
            max_mass[ii] = 20.0
        ii += 1

    return min_mass[0], max_mass[0], min_mass[1], max_mass[1]


def idl_tabulate(x, f, p=5):
    """Function that replicates the IDL int_tabulated function
    which performs a p-point integration on a tabulated set of data

    Parameters
    ----------
    x : array
        tabulated x-value data
    f : array
        tabulated f-value data, same size as x
    p : int
        number of chunks to divide tabulated data into
        Default: 5

    Returns
    -------
    ret : float
        Integration result
    """

    def newton_cotes(x, f):
        if x.shape[0] < 2:
            return 0
        rn = (x.shape[0] - 1) * (x - x[0]) / (x[-1] - x[0])
        weights = scipy.integrate.newton_cotes(rn)[0]
        return (x[-1] - x[0]) / (x.shape[0] - 1) * np.dot(weights, f)

    ret = 0
    for idx in range(0, x.shape[0], p - 1):
        ret += newton_cotes(x[idx: idx + p], f[idx: idx + p])
    return ret


def rndm(a, b, g, size):
    r"""Power-law generator for pdf(x)\propto x^{g} for a<=x<=b

    Parameters
    ----------
    a : float
        Minimum of range for power law
    b : float
        Maximum of range for power law
    g : float
        Index for power law
    size : int
        Number of data points to draw

    Returns
    -------
    power : array
        Array of data sampled from power law distribution with params
        fixed by inputs
    """

    r = np.random.random(size=size)
    ag, bg = a ** (g + 1), b ** (g + 1)
    return (ag + (bg - ag) * r) ** (1.0 / (g + 1))


def param_transform(dat):
    """Transforms a data set to limits between zero and one
    Leaves some wiggle room on the edges of the data set

    Parameters
    ----------
    dat : array
        array of data to transform between 0 and 1

    Returns
    -------
    datTransformed : array
        array of data with limits between 0 and 1
    """

    datMax = max(dat)
    datMin = min(dat)
    datZeroed = dat - datMin

    datTransformed = datZeroed / ((datMax - datMin))
    if np.max(datTransformed) == 1.0:
        datTransformed[datTransformed == 1.0] = 1 - 1e-6
    if np.min(datTransformed) == 0.0:
        datTransformed[datTransformed == 0.0] = 1e-6
    return datTransformed


def dat_transform(dat, dat_list):
    """Transform a data set to have limits between zero and one using
    param_transform, then transform to log space

    Parameters
    ----------
    dat " DataFrame
        Data to transform to eventually perform KDE
    dat_list : list
        List of DataFrame columns to include in transformation

    Returns
    -------
    dat_trans : array
        Transformed data for columns in dat_list
    """

    dat_trans = []
    for column in dat_list:
        dat_trans.append(ss.logit(param_transform(dat[column])))
    dat_trans = np.vstack([dat_trans])

    return dat_trans


def dat_un_transform(dat_sample, dat_set, dat_list):
    """Un-transform data that was transformed in dat_transform

    Parameters
    ----------
    dat_sample : array
        Data sampled from kde generated with transformed data
    dat_set : DataFrame
        Un-transformed data (same as dat in dat_transform)
    dat_list : list
        List of DataFrame columns to include in transformation

    Returns
    -------
    dat : array
        Array of data sampled from kde that is transformed back to
        bounds of the un-transformed data set the kde is generated from
    """
    dat = []

    dat_exp = ss.expit(dat_sample)
    for ii, column in zip(range(len(dat_list)), dat_list):
        dat_untrans = dat_exp[ii, :] * (
            max(dat_set[column]) - min(dat_set[column])
        ) + min(dat_set[column])
        dat.append(dat_untrans)
    dat = np.vstack(dat)
    return dat


def knuth_bw_selector(dat_list):
    """Selects the kde bandwidth using Knuth's rule implemented in Astropy
    If Knuth's rule raises error, Scott's rule is used

    Parameters
    ----------
    dat_list : list
        List of data arrays that will be used to generate a kde

    Returns
    -------
    bw_min : float
        Minimum of bandwidths for all of the data arrays in dat_list
    """

    bw_list = []
    for dat in dat_list:
        try:
            bw = astrostats.knuth_bin_width(dat)
        except Exception:
            print("Using Scott Rule!!")
            bw = astrostats.scott_bin_width(dat)
        bw_list.append(bw)
    return np.mean(bw_list)


def get_Z_from_FeH(FeH, Z_sun=0.02):
    """
    Converts from FeH to Z under the assumption that
    all stars have the same abundance as the sun
    
    Parameters
    ----------
    FeH : array
        Fe/H values to convert
    Z_sun : float
        solar metallicity
    
    Returns
    -------
    Z : array
        metallicities corresponding to Fe/H
    """
    Z = 10**(FeH + np.log10(Z_sun))
    return Z


def get_FeH_from_Z(Z, Z_sun=0.02):
    """
    Converts from Z to FeH under the assumption that
    all stars have the same abundance as the sun
    
    Parameters
    ----------
    Z : array
        metallicities to convert to Fe/H
    Z_sun : float
        solar metallicity
    
    Returns
    -------
    FeH : array
        Fe/H corresponding to metallicities
    """
    FeH = np.log10(Z) - np.log10(Z_sun)
    return FeH


def get_binfrac_of_Z(Z):
    '''
    Calculates the theoretical binary fraction as a 
    function of metallicity. Following Moe+2019
    
    Parameters
    ----------
    Z : array
        metallicity Z values
    
    Returns
    -------
    binfrac : array
        binary fraction values
    '''
    FeH = get_FeH_from_Z(Z)
    FeH_low = FeH[np.where(FeH<=-1.0)]
    FeH_high = FeH[np.where(FeH>-1.0)]
    binfrac_low = -0.0648 * FeH_low + 0.3356
    binfrac_high = -0.1977 * FeH_high + 0.2025
    binfrac = np.append(binfrac_low, binfrac_high)
    return binfrac


def get_porb_norm(Z, close_logP=4.0, wide_logP=6.0, binfrac_tot_solar=0.66, Z_sun=0.02):
    '''Returns normalization constants to produce log normals consistent with Fig 19 of Moe+19
    for the orbital period distribution
                
    Parameters
    ----------
    Z : array
        metallicity values
    close_logP : float
        divding line beween close and intermediate orbits
    wide_logP : float
        dividing line between intermediate and wide orbits
    binfrac_tot : float
        integrated total binary fraction at solar metallicity
                
    Returns
    -------
    norm_wide : float
        normalization factor for kde for wide binaries
    norm_close : float
        normalization factor for kde for wide binaries
    '''
    from scipy.stats import norm
    from scipy.integrate import trapz
    from scipy.interpolate import interp1d
    
    # fix to values used in Moe+19
    logP_lo_lim=0
    logP_hi_lim=9
    log_P = np.linspace(logP_lo_lim, logP_hi_lim, 10000)
    
    logP_pdf = norm.pdf(log_P, loc=4.9, scale=2.3)
    
    # set up the wide binary fraction inflection point
    norm_wide = binfrac_tot_solar/trapz(logP_pdf, log_P)
    
    # set up the close binary fraction inflection point
    FeHclose = np.linspace(-3.0, 0.5, 100)
    fclose = -0.0648 * FeHclose + 0.3356
    fclose[FeHclose > -1.0] = -0.1977 * FeHclose[FeHclose > -1.0] + 0.2025
    Zclose = get_Z_from_FeH(FeHclose, Z_sun=Z_sun)
    
    fclose_interp = interp1d(Zclose, fclose)
    
    fclose_Z = fclose_interp(Z)
    norm_close = fclose_Z/trapz(logP_pdf[log_P < close_logP], log_P[log_P < close_logP])
    
    return norm_wide, norm_close


def get_met_dep_binfrac(met):
    '''Returns a population-wide binary fraction consistent with
    Moe+19 based on the supplied metallicity

    Parameters
    ----------
    met : float
        metallicity of the population

    Returns
    -------
    binfrac : float
        binary fraction of the population based on metallicity

    '''
    logP_hi_lim = 9
    logP_lo_lim = 0
    wide_logP = 6
    close_logP = 4
    neval = 5000

    from scipy.interpolate import interp1d
    from scipy.integrate import trapz
    from scipy.stats import norm

    norm_wide, norm_close = get_porb_norm(met)
    prob_wide = norm.pdf(np.linspace(wide_logP, logP_hi_lim, neval), loc=4.9, scale=2.3)*norm_wide
    prob_close = norm.pdf(np.linspace(logP_lo_lim, close_logP, neval), loc=4.9, scale=2.3)*norm_close
    slope = -(prob_close[-1] - prob_wide[0]) / (wide_logP - close_logP)
    prob_intermediate = slope * (np.linspace(close_logP, wide_logP, neval) - close_logP) + prob_close[-1]
    prob_interp_int = interp1d(np.linspace(close_logP, wide_logP, neval), prob_intermediate)

    x_dat = np.hstack([np.linspace(logP_lo_lim, close_logP, neval),
                       np.linspace(close_logP, wide_logP, neval),
                       np.linspace(wide_logP, logP_hi_lim, neval),])
    y_dat = np.hstack([prob_close, prob_interp_int(np.linspace(close_logP, wide_logP, neval)), prob_wide])

    binfrac = trapz(y_dat, x_dat)/0.66 * 0.5

    return float(np.round(binfrac, 2))

def error_check(BSEDict, filters=None, convergence=None, sampling=None):
    """Checks that values in BSEDict, filters, and convergence are viable"""
    if not isinstance(BSEDict, dict):
        raise ValueError("BSE flags must be supplied via a dictionary")

    if filters is not None:
        if not isinstance(filters, dict):
            raise ValueError("Filters criteria must be supplied via a dictionary")
        for option in ["binary_state", "timestep_conditions"]:
            if option not in filters.keys():
                raise ValueError(
                    "Inifile section filters must have option {0} supplied".format(
                        option
                    )
                )

    if convergence is not None:
        if not isinstance(convergence, dict):
            raise ValueError("Convergence criteria must be supplied via a dictionary")
        for option in [
            "pop_select",
            "convergence_params",
            "convergence_limits",
            "match",
            "apply_convergence_limits",
        ]:
            if option not in convergence.keys():
                raise ValueError(
                    "Inifile section convergence must have option {0} supplied".format(
                        option
                    )
                )

    if sampling is not None:
        if not isinstance(sampling, dict):
            raise ValueError("Sampling criteria must be supplied via a dictionary")
        for option in ["sampling_method", "SF_start", "SF_duration", "metallicity", "keep_singles"]:
            if option not in sampling.keys():
                raise ValueError(
                    "Inifile section sampling must have option {0} supplied".format(
                        option
                    )
                )
        if ("qmin" not in sampling.keys()) & ("m2_min" not in sampling.keys()) & (sampling["sampling_method"] == 'independent'):
            raise ValueError("You have not specified qmin or m2_min. At least one of these must be specified.")
    # filters
    if filters is not None:
        flag = "binary_state"
        if any(x not in [0, 1, 2] for x in filters[flag]):
            raise ValueError(
                "{0} needs to be a subset of [0,1,2] (you set it to {1})".format(
                    flag, filters[flag]
                )
            )
        flag = "timestep_conditions"
        if (type(filters[flag]) != str) and (type(filters[flag]) != list):
            raise ValueError(
                "{0} needs to either be a string like 'dtp=None' or a list of conditions like [['binstate==0', 'dtp=1.0']] (you set it to {1})".format(
                    flag, filters[flag]
                )
            )

    # convergence
    if convergence is not None:
        flag = "convergence_limits"
        if convergence[flag]:
            for item, key in zip(convergence.items(), convergence.keys()):
                if len(item) != 2:
                    raise ValueError(
                        "The value for key '{0:s}' needs to be a list of length 2, it is length: {1:i}".format(
                            key, len(item)
                        )
                    )
        flag = "pop_select"
        if not convergence[flag] in [
            "formation",
            "1_SN",
            "2_SN",
            "disruption",
            "final_state",
            "XRB_form",
        ]:
            raise ValueError(
                "{0} needs to be in the list: ['formation', '1_SN', '2_SN', 'disruption', 'final_state', 'XRB_form'] "
                "(you set it to {1})".format(
                                             flag, convergence[flag]
                                            )
            )

        flag = "match"
        if not isinstance(convergence[flag], float):
            raise ValueError(
                "{0} must be a float (you set it to {1})".format(
                    flag, convergence[flag]
                )
            )

        flag = "convergence_params"
        acceptable_convergence_params = [
            "mass_1",
            "mass_2",
            "sep",
            "porb",
            "ecc",
            "massc_1",
            "massc_2",
            "rad_1",
            "rad_2",
        ]
        for param in convergence[flag]:
            if param not in acceptable_convergence_params:
                raise ValueError(
                    "Supplied convergence parameter {0} is not in list of "
                    "acceptable convergence parameters {1}".format(
                        param, acceptable_convergence_params
                    )
                )

        flag = "convergence_limits"
        if type(convergence[flag]) != dict:
            raise ValueError(
                "Supplied convergence limits must be passed as a dict "
                "(you passed type {0})".format(type(convergence[flag]))
                )

        for key in convergence[flag].keys():
            if key not in convergence["convergence_params"]:
                raise ValueError(
                    "Supplied convergence limits must correspond to already "
                    "supplied convergence_params. The supplied convergence_params "
                    "are {0}, while you supplied {1}".format(
                        convergence["convergence_params"], key)
                    )
        flag = "apply_convergence_limits"
        if type(convergence[flag]) != bool:
            raise ValueError(
                "apply_convergence_limits must be either True or False, "
                "you supplied {}".format(
                    convergence[flag])
                )  
 
    # sampling
    if sampling is not None:
        flag = "sampling_method"
        acceptable_sampling = ["multidim", "independent"]
        if sampling[flag] not in acceptable_sampling:
            raise ValueError(
                "sampling_method must be one of {0} you supplied {1}.".format(
                    acceptable_sampling, sampling[flag]
                )
            )

        flag = "metallicity"
        if not isinstance(sampling[flag], float):
            raise ValueError(
                "{0} must be a float (you set it to {1})".format(flag, sampling[flag])
            )
        if sampling[flag] <= 0:
            raise ValueError(
                "{0} needs to be greater than or equal to 0 (you set it to {1})".format(
                    flag, sampling[flag]
                )
            )

    # BSEDict
    flag = "dtp"
    if flag in BSEDict.keys():
        if BSEDict[flag] < 0:
            raise ValueError(
                "'{0:s}' needs to be greater than or equal to 0 (you set it to '{1:0.2f}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "pts1"
    if flag in BSEDict.keys():
        if BSEDict[flag] <= 0:
            raise ValueError(
                "'{0:s}' needs to be greater than 0 (you set it to '{1:0.2f}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "pts2"
    if flag in BSEDict.keys():
        if BSEDict[flag] <= 0:
            raise ValueError(
                "'{0:s}' needs to be greater than 0 (you set it to '{1:0.2f}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "pts3"
    if flag in BSEDict.keys():
        if BSEDict[flag] <= 0:
            raise ValueError(
                "'{0:s}' needs to be greater than 0 (you set it to '{1:0.2f}')".format(
                    flag, BSEDict[flag]
                )
            )

    flag = "zsun"
    if flag in BSEDict.keys():
        #if BSEDict[flag] != 0.019:
        #    warnings.warn(
        #        f"'{flag:s}' is set to a different value than assumed in the mlwind "
        #        f"prescriptions (you set it to '{BSEDict[flag]:0.3f}' and in mlwind, zsun_wind=0.019)"
        #    )
        if BSEDict[flag] <= 0:
            raise ValueError(
                "'{0:s}' needs to be greater than 0 (you set it to '{1:0.2f}')".format(
                    flag, BSEDict[flag]
                )
            )

    flag = "windflag"
    if flag in BSEDict.keys():
        if BSEDict[flag] not in [0, 1, 2, 3]:
            raise ValueError(
                "'{0:s}' needs to be set to either 0, 1, 2, or 3 (you set it to '{1:d}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "rtmsflag"
    if flag in BSEDict.keys():
        if BSEDict[flag] not in [0, 1, 2]:
            raise ValueError(
                "'{0:s}' needs to be set to either 0, 1, or 2 (you set it to '{1:d}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "eddlimflag"
    if flag in BSEDict.keys():
        if BSEDict[flag] not in [0, 1]:
            raise ValueError(
                "'{0:s}' needs to be set to either 0 or 1 (you set it to '{1:d}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "neta"
    if flag in BSEDict.keys():
        if BSEDict[flag] <= 0:
            raise ValueError(
                "'{0:s}' needs to be greater than 0 (you set it to '{1:0.2f}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "bwind"
    if flag in BSEDict.keys():
        if BSEDict[flag] < 0:
            raise ValueError(
                "'{0:s}' needs to be greater or equal to 0 (you set it to '{1:0.2f}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "hewind"
    if flag in BSEDict.keys():
        if (BSEDict[flag] < 0) or (BSEDict[flag] > 1):
            raise ValueError(
                "'{0:s}' needs to be between 0 and 1 (you set it to '{1:0.2f}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "beta"
    # --- all numbers are valid
    flag = "xi"
    if flag in BSEDict.keys():
        if (BSEDict[flag] < 0) or (BSEDict[flag] > 1):
            raise ValueError(
                "'{0:s}' needs to be between 0 and 1 (you set it to '{1:0.2f}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "acc2"
    if flag in BSEDict.keys():
        if BSEDict[flag] < 0:
            raise ValueError(
                "'{0:s}' needs to be greater or equal to 0 (you set it to '{1:0.2f}')".format(
                    flag, BSEDict[flag]
                )
            )

    flag = "alpha1"
    if flag in BSEDict.keys():
        if BSEDict[flag] <= 0:
            raise ValueError(
                "'{0:s}' needs to be greater than 0 (you set it to '{1:0.2f}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "lambdaf"
    # --- all numbers are valid
    flag = "ceflag"
    if flag in BSEDict.keys():
        if BSEDict[flag] not in [0, 1]:
            raise ValueError(
                "'{0:s}' needs to be set to either 0 or 1 (you set it to '{1:d}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "cekickflag"
    if flag in BSEDict.keys():
        if BSEDict[flag] not in [0, 1, 2]:
            raise ValueError(
                "'{0:s}' needs to be set to either 0, 1, or 2 (you set it to '{1:d}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "cemergeflag"
    if flag in BSEDict.keys():
        if BSEDict[flag] not in [0, 1]:
            raise ValueError(
                "'{0:s}' needs to be set to either 0 or 1 (you set it to '{1:d}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "cehestarflag"
    if flag in BSEDict.keys():
        if BSEDict[flag] not in [0, 1, 2]:
            raise ValueError(
                "'{0:s}' needs to be set to either 0, 1, or 2 (you set it to '{1:d}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "grflag"
    if flag in BSEDict.keys():
        if BSEDict[flag] not in [0, 1]:
            raise ValueError(
                "'{0:s}' needs to be set to either 0 or 1 (you set it to '{1:d}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "qcflag"
    if flag in BSEDict.keys():
        if BSEDict[flag] not in [0, 1, 2, 3, 4, 5]:
            raise ValueError(
                "'{0:s}' needs to be set to 0, 1, 2, 3, 4, or 5 (you set it to '{1:0.2f}')".format(
                    flag, BSEDict[flag]
                )
            )

    flag = "qcrit_array"
    if flag in BSEDict.keys():
        if any(x < 0.0 for x in BSEDict[flag]):
            raise ValueError(
                "'{0:s}' values must be greater than or equal to zero (you set them to '[{1:d}]')".format(
                    flag, *BSEDict[flag]
                )
            )
        if len(BSEDict[flag]) != 16:
            raise ValueError(
                "'{0:s}' must be supplied 16 values (you supplied '{1:d}')".format(
                    flag, len(BSEDict[flag])
                )
            )

    flag = "kickflag"
    if flag in BSEDict.keys():
        if BSEDict[flag] not in [0, -1, -2, -3]:
            raise ValueError(
                "'{0:s}' needs to be set to either 0, -1, -2, or -3 (you set it to '{1:d}')".format(
                    flag, BSEDict[flag]
                )
            )
        if (BSEDict[flag] in [-1, -2]):
            if (BSEDict['ecsn'] != 2.25) or (BSEDict['ecsn_mlow'] != 1.6):
                warnings.warn("You have chosen a kick flag that assumes compact object formation "
                              "according to Giacobbo & Mapelli 2020, but supplied electron "
                              "capture SN (ECSN) flags that are inconsistent with this study. "
                              "To maintain consistency, COSMIC will update your "
                              "ECSN flags to be ecsn=2.25 and ecsn_mlow=1.6")
                BSEDict['ecsn'] = 2.25
                BSEDict['ecsn_mlow'] = 1.6
    flag = "sigma"
    if flag in BSEDict.keys():
        if BSEDict[flag] < 0:
            raise ValueError(
                "'{0:s}' needs to be greater or equal to 0 (you set it to '{1:0.2f}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "bhflag"
    if flag in BSEDict.keys():
        if BSEDict[flag] not in [0, 1, 2, 3]:
            raise ValueError(
                "'{0:s}' needs to be set to either 0, 1, 2, or 3 (you set it to '{1:d}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "ecsn"
    if flag in BSEDict.keys():
        if BSEDict[flag] < 0:
            raise ValueError(
                "'{0:s}' needs to be greater or equal to 0 (you set it to '{1:0.2f}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "ecsn_mlow"
    if flag in BSEDict.keys():
        if (BSEDict[flag] > BSEDict["ecsn"]) or (BSEDict[flag] < 0.0):
            raise ValueError(
                "'{0:s}' needs to be less than 'ecsn', and must be greater than or equal to 0 "
                "(you set it to '{1:0.2f}')".format(
                                                    flag, BSEDict[flag]
                                                   )
            )
    flag = "sigmadiv"
    if flag in BSEDict.keys():
        if BSEDict[flag] == 0:
            raise ValueError(
                "'{0:s}' must be positive or negative (you set it to '{1:0.2f}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "aic"
    if flag in BSEDict.keys():
        if BSEDict[flag] not in [0, 1]:
            raise ValueError(
                "'{0:s}' needs to be set to either 0 or 1 (you set it to '{1:d}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "ussn"
    if flag in BSEDict.keys():
        if BSEDict[flag] not in [0, 1]:
            raise ValueError(
                "'{0:s}' needs to be set to either 0 or 1 (you set it to '{1:d}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "pisn"
    if flag in BSEDict.keys():
        if not (
            (BSEDict[flag] >= 0)
            or (BSEDict[flag] == -1)
            or (BSEDict[flag] == -2)
            or (BSEDict[flag] == -3)
        ):
            raise ValueError(
                "'{0:s}' needs to be set to either 0, greater than 0 or equal to -1, -2, or -3 "
                "(you set it to '{1:0.2f}')".format(
                                                    flag, BSEDict[flag]
                                                   )
            )
    flag = "bhsigmafrac"
    if flag in BSEDict.keys():
        if (BSEDict[flag] <= 0) or (BSEDict[flag] > 1):
            raise ValueError(
                "'{0:s}' needs to be greater than 0 and less than or equal to 1 (you set it to '{1:0.2f}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "polar_kick_angle"
    if flag in BSEDict.keys():
        if (BSEDict[flag] < 0) or (BSEDict[flag] > 90):
            raise ValueError(
                "'{0:s}' needs to be within the allowed range of [0,90] (you set it to '{1:0.2f}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "natal_kick_array"
    if flag in BSEDict.keys():
        if np.array(BSEDict[flag]).shape != (2, 5):
            raise ValueError(
                "'{0:s}' must have shape (2,5) (you supplied list, or array with shape '{1}')".format(
                    flag, np.array(BSEDict[flag]).shape
                )
            )

    flag = "remnantflag"
    if flag in BSEDict.keys():
        if BSEDict[flag] not in [0, 1, 2, 3, 4]:
            raise ValueError(
                "'{0:s}' needs to be set to either 0, 1, 2, 3, or 4 (you set it to '{1:d}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "mxns"
    if flag in BSEDict.keys():
        if BSEDict[flag] <= 0:
            raise ValueError(
                "'{0:s}' needs to be greater than 0 (you set it to '{1:0.2f}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "rembar_massloss"
    if flag in BSEDict.keys():
        if BSEDict[flag] < -1:
            raise ValueError(
                "'{0:s}' needs to be between [-1,0] or greater than 0 (you set it to '{1:0.2f}')".format(
                    flag, BSEDict[flag]
                )
            )

    flag = "eddfac"
    if flag in BSEDict.keys():
        if BSEDict[flag] < 0:
            raise ValueError(
                "'{0:s}' needs to be greater or equal to 0 (you set it to '{1:0.2f}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "gamma"
    if flag in BSEDict.keys():
        if (BSEDict[flag] <= 0) and (BSEDict[flag] != -1) and (BSEDict[flag] != -2):
            raise ValueError(
                "'{0:s}' needs to either be set to -2, -1, or a positive number (you set it to '{1:0.2f}')".format(
                    flag, BSEDict[flag]
                )
            )

    flag = "tflag"
    if flag in BSEDict.keys():
        if BSEDict[flag] not in [0, 1]:
            raise ValueError(
                "'{0:s}' needs to be set to either 0 or 1 (you set it to '{1:d}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "ifflag"
    if flag in BSEDict.keys():
        if BSEDict[flag] < 0:
            raise ValueError(
                "'{0:s}' needs to be greater or equal to 0 (you set it to '{1:0.2f}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "wdflag"
    if flag in BSEDict.keys():
        if BSEDict[flag] < 0:
            raise ValueError(
                "'{0:s}' needs to be greater or equal to 0 (you set it to '{1:0.2f}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "epsnov"
    if flag in BSEDict.keys():
        if (BSEDict[flag] < 0) or (BSEDict[flag] > 1):
            raise ValueError(
                "'{0:s}' needs to be between 0 and 1 (you set it to '{1:0.2f}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "bhspinflag"
    if flag in BSEDict.keys():
        if BSEDict[flag] not in [0, 1, 2]:
            raise ValueError(
                "'{0:s}' needs to be set to 0, 1, or 2 (you set it to '{1:0.2f}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "bhspinmag"
    if flag in BSEDict.keys():
        if (BSEDict[flag] < 0) or (BSEDict[flag] > 1):
            raise ValueError(
                "'{0:s}' needs to be between 0 and 1 (you set it to '{1:0.2f}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "bconst"
    if flag in BSEDict.keys():
        if BSEDict[flag] <= 0:
            raise ValueError(
                "'{0:s}' needs to be greater than 0 (you set it to '{1:0.2f}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "ck"
    if flag in BSEDict.keys():
        if BSEDict[flag] <= 0:
            raise ValueError(
                "'{0:s}' needs to be greater than 0 (you set it to '{1:0.2f}')".format(
                    flag, BSEDict[flag]
                )
            )

    flag = "fprimc_array"
    if flag in BSEDict.keys():
        if any(x < 0.0 for x in BSEDict[flag]):
            raise ValueError(
                "'{0:s}' values must be greater than or equal to zero (you set them to '[{1:d}]')".format(
                    flag, *BSEDict[flag]
                )
            )
        if len(BSEDict[flag]) != 16:
            raise ValueError(
                "'{0:s}' must be supplied 16 values (you supplied '{1:d}')".format(
                    flag, len(BSEDict[flag])
                )
            )
    flag = "rejuv_fac"
    if flag in BSEDict.keys():
        if (BSEDict[flag] > 1.0) or (BSEDict[flag] < 0.0):
            raise ValueError(
                "'{0:s}' must be between 0 and 1 (you set it to '[{1:d}]')".format(
                    flag, *BSEDict[flag]
                )
            )
    flag = "rejuv_flag"
    if flag in BSEDict.keys():
        if BSEDict[flag] not in [0, 1]:
            raise ValueError(
                "'{0:s}' needs to be set to 0 or 1 (you set it to '{1:0.2f}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "htpmb"
    if flag in BSEDict.keys():
        if BSEDict[flag] not in [0, 1]:
            raise ValueError(
                "'{0:s}' needs to be set to 0 or 1 (you set it to '{1:0.2f}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "bdecayfac"
    if flag in BSEDict.keys():
        if BSEDict[flag] not in [0, 1]:
            raise ValueError(
                "'{0:s}' needs to be set to 0 or 1 (you set it to '{1:0.2f}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "ST_cr"
    if flag in BSEDict.keys():
        if BSEDict[flag] not in [0, 1]:
            raise ValueError(
                "'{0:s}' needs to be set to 0 or 1 (you set it to '{1:0.2f}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "ST_tide"
    if flag in BSEDict.keys():
        if BSEDict[flag] not in [0, 1]:
            raise ValueError(
                "'{0:s}' needs to be set to 0 or 1 (you set it to '{1:0.2f}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "don_lim"
    if flag in BSEDict.keys():
        if BSEDict[flag] not in [-1, -2]:
            raise ValueError(
                "'{0:s}' needs to be set to -1 or -2 (you set it to '{1:0.2f}')".format(
                    flag, BSEDict[flag]
                )
            )
    flag = "acc_lim"
    if flag in BSEDict.keys():
        if BSEDict[flag] not in [-1, -2, -3, -4]:
            if BSEDict[flag] < 0.0:
                raise ValueError(
                    "'{0:s}' needs to be set to -1, -2, -3, -4 or be >=0 (you set it to '{1:0.2f}')".format(
                        flag, BSEDict[flag]
                    )
                )

    return


def check_initial_conditions(initial_binary_table):
    """Checks initial conditions and reports warnings

    Only warning provided right now is if star begins in Roche lobe
    overflow
    """

    def rzamsf(m):
        """A function to evaluate Rzams
        ( from Tout et al., 1996, MNRAS, 281, 257 ).
        """
        mx = np.sqrt(m)
        rzams = (
            (a[7] * m ** 2 + a[8] * m ** 6) * mx
            + a[9] * m ** 11
            + (a[10] + a[11] * mx) * m ** 19
        ) / (a[12] + a[13] * m ** 2 + (a[14] * m ** 8 + m ** 18 + a[15] * m ** 19) * mx)

        return rzams

    z = np.asarray(initial_binary_table["metallicity"])
    zpars, a = zcnsts(z)

    mass1 = np.asarray(initial_binary_table["mass_1"])
    mass2 = np.asarray(initial_binary_table["mass_2"])

    if np.all(mass2 == 0.0):
        return
    else:
        rzams1 = rzamsf(mass1)
        rzams2 = rzamsf(mass2)

        # assume some time step in order to calculate sep
        yeardy = 365.24
        aursun = 214.95
        tb = np.asarray(initial_binary_table["porb"]) / yeardy
        sep = aursun * (tb * tb * (mass1 + mass2)) ** (1.0 / 3.0)

        rol1 = calc_Roche_radius(mass1, mass2, sep)
        rol2 = calc_Roche_radius(mass2, mass1, sep)

        # check for a ZAMS that starts in RFOL
        mask = ((np.array(initial_binary_table["kstar_1"]) == 1) & (rzams1 >= rol1)) | (
            (initial_binary_table["kstar_2"] == 1) & (rzams2 >= rol2)
        )
        if mask.any():
            warnings.warn(
                "At least one of your initial binaries is starting in Roche Lobe Overflow:\n{0}".format(
                    initial_binary_table[mask]
                )
            )

        return


def convert_kstar_evol_type(bpp):
    """Provides way to convert integer values to their string counterpart

    The underlying fortran code relies on integers to indicate
    things like the evoltuionary stage of the star as well as
    key moments in its evolutionary track. If you pass the
    data frame returned from running

        ```Evolve.evolve```

    you can convert the columns with these integer proxies
    to their true astrophysical meaning.
    """
    kstar_int_to_string_dict = {
        0: "Main Sequence (MS), < 0.7 M⊙",
        1: "MS, > 0.7 M⊙",
        2: "Hertzsprung Gap",
        3: "First Giant Branch",
        4: "Core Helium Burning",
        5: "Early Asymptotic Giant Branch (AGB)",
        6: "Thermally Pulsing AGB",
        7: "Naked Helium Star MS",
        8: "Naked Helium Star Hertzsprung Gap",
        9: "Naked Helium Star Giant Branch",
        10: "Helium White Dwarf",
        11: "Carbon/Oxygen White Dwarf",
        12: "Oxygen/Neon White Dwarf",
        13: "Neutron Star",
        14: "Black Hole",
        15: "Massless Remnant",
    }

    kstar_string_to_int_dict = {v: k for k, v in kstar_int_to_string_dict.items()}

    evolve_type_int_to_string_dict = {
        1: "initial state",
        2: "kstar change",
        3: "begin Roche lobe overflow",
        4: "end Roche lobe overlow",
        5: "contact",
        6: "coalescence",
        7: "begin common envelope",
        8: "end common envelope",
        9: "no remnant leftover",
        10: "max evolution time",
        11: "binary disruption",
        12: "begin symbiotic phase",
        13: "end symbiotic phase",
        14: "blue straggler",
        15: "supernova of primary",
        16: "supernova of secondary",
       100: "RLOF interpolation timeout error"
    }

    evolve_type_string_to_int_dict = {
        v: k for k, v in evolve_type_int_to_string_dict.items()
    }

    if bpp.kstar_1.dtype in [int, float]:
        # convert from integer to string
        bpp["kstar_1"] = bpp["kstar_1"].astype(int)
        bpp["kstar_1"] = bpp["kstar_1"].apply(lambda x: kstar_int_to_string_dict[x])
    else:
        # convert from string to integer
        bpp["kstar_1"] = bpp["kstar_1"].apply(lambda x: kstar_string_to_int_dict[x])

    if bpp.kstar_2.dtype in [int, float]:
        # convert from integer to string
        bpp["kstar_2"] = bpp["kstar_2"].astype(int)
        bpp["kstar_2"] = bpp["kstar_2"].apply(lambda x: kstar_int_to_string_dict[x])
    else:
        # convert from string to integer
        bpp["kstar_2"] = bpp["kstar_2"].apply(lambda x: kstar_string_to_int_dict[x])

    if bpp.evol_type.dtype in [int, float]:
        # convert from integer to string
        bpp["evol_type"] = bpp["evol_type"].astype(int)
        bpp["evol_type"] = bpp["evol_type"].apply(
            lambda x: evolve_type_int_to_string_dict[x]
        )
    else:
        # convert from string to integer
        bpp["evol_type"] = bpp["evol_type"].apply(
            lambda x: evolve_type_string_to_int_dict[x]
        )

    return bpp


def parse_inifile(inifile):
    """Provides a method for parsing the inifile and returning dicts of each section"""
    if inifile is None:
        raise ValueError("Please supply an inifile")
    elif not os.path.isfile(inifile):
        raise ValueError("inifile supplied does not exist")

    binOps = {
        ast.Add: operator.add,
        ast.Sub: operator.sub,
        ast.Mult: operator.mul,
        ast.Div: operator.truediv,
        ast.Mod: operator.mod,
    }

    def arithmetic_eval(s):
        """Allows us to control how the strings from the inifile get parses"""
        node = ast.parse(s, mode="eval")

        def _eval(node):
            """Different strings receive different evaluation"""
            if isinstance(node, ast.Expression):
                return _eval(node.body)
            elif isinstance(node, ast.Str):
                return node.s
            elif isinstance(node, ast.Num):
                return node.n
            elif isinstance(node, ast.BinOp):
                return binOps[type(node.op)](_eval(node.left), _eval(node.right))
            elif isinstance(node, ast.List):
                return [_eval(x) for x in node.elts]
            elif isinstance(node, ast.Name):
                result = VariableKey(item=node)
                constants_lookup = {
                    "True": True,
                    "False": False,
                    "None": None,
                }
                value = constants_lookup.get(
                    result.name,
                    result,
                )
                if type(value) == VariableKey:
                    # return regular string
                    return value.name
                else:
                    # return special string like True or False
                    return value
            elif isinstance(node, ast.NameConstant):
                # None, True, False are nameconstants in python3, but names in 2
                return node.value
            else:
                raise Exception("Unsupported type {}".format(node))

        return _eval(node.body)

    # ---- Create configuration-file-parser object and read parameters file.
    cp = ConfigParser()
    cp.optionxform = str
    cp.read(inifile)

    # ---- Read needed variables from the inifile
    dictionary = {}
    for section in cp.sections():
        # for cosmic we skip any CMC stuff
        # if "cmc" is a section in the ini file, then we can optionally skip the 
        # COSMIC population sections (or not, if they exist)
        if section == "cmc":
            if "rand_seed" not in dictionary.keys():
                dictionary["rand_seed"] = {}
                dictionary["rand_seed"]["seed"] = 0
            if "filters" not in dictionary.keys():
                dictionary["filters"] = 0
            if "convergence" not in dictionary.keys():
                dictionary["convergence"] = 0
            if "sampling" not in dictionary.keys():
                dictionary["sampling"] = 0
            continue
        dictionary[section] = {}
        for option in cp.options(section):
            opt = cp.get(section, option)
            if "\n" in opt:
                raise ValueError("We have detected an error in your inifile. A parameter was read in with the following "
                                 "value: {0}. Likely, you have an unexpected syntax, such as a space before an parameter/option (i.e. "
                                 "the parameter must be flush to the far left of the file".format(opt))
            try:
                dictionary[section][option] = arithmetic_eval(opt)
            except Exception:
                dictionary[section][option] = json.loads(opt)
            finally:
                if option not in dictionary[section].keys():
                    raise ValueError("We have detected an error in your inifile. The folloiwng parameter failed to be read correctly: {0}".format(option))

    BSEDict = dictionary["bse"]
    seed_int = int(dictionary["rand_seed"]["seed"])
    filters = dictionary["filters"]
    convergence = dictionary["convergence"]
    sampling = dictionary["sampling"]

    return BSEDict, seed_int, filters, convergence, sampling


class VariableKey(object):
    """
    A dictionary key which is a variable.
    @ivar item: The variable AST object.
    """

    def __init__(self, item):
        self.name = item.id

    def __eq__(self, compare):
        return compare.__class__ == self.__class__ and compare.name == self.name

    def __hash__(self):
        return hash(self.name)



"""Unit test for cosmic
"""

__author__ = 'Katie Breivik <katie.breivik@gmail.com>'

import os
import unittest
import numpy as np
import scipy.integrate
import pandas as pd

from .. import Match

np.random.seed(2)
sample = np.random.uniform(0,1,500)

MATCH_TEST = np.log10(1-0.99636185870762237007)

class TestMatch(unittest.TestCase):
    """`TestCase` for the match method
    """
    def test_Match(self):
        # test the match by dividing the sample into to sub samples
        # one containing half of the set and one containing the full set

        dataCm = [sample[:int(len(sample)/2)], sample]
        match, bin_width = Match.match(dataCm)
        self.assertAlmostEqual(match,MATCH_TEST)





#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) Katie Breivik (2017-2020)
#
# This file is part of the cosmic python package.
#
# cosmic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cosmic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cosmic.  If not, see <http://www.gnu.org/licenses/>.

"""Setup the cosmic package
"""

from __future__ import print_function

import glob
import os.path
import sys

from setuptools import find_packages
from distutils.command.sdist import sdist

# set basic metadata
PACKAGENAME = "cosmic"
DISTNAME = "cosmic-popsynth"
AUTHOR = "Katie Breivik"
AUTHOR_EMAIL = "katie.breivik@gmail.com"
LICENSE = "GPLv3"

cmdclass = {}

# -- versioning ---------------------------------------------------------------
import re
VERSIONFILE = "cosmic/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))

# -- documentation ------------------------------------------------------------

# import sphinx commands
try:
    from sphinx.setup_command import BuildDoc
except ImportError:
    pass
else:
    cmdclass["build_sphinx"] = BuildDoc

cmdclass["sdist"] = sdist

# read description
with open("README.md", "rb") as f:
    long_description = f.read().decode().strip()

# -- dependencies -------------------------------------------------------------

setup_requires = [
    "numpy",
]

if 'test' in sys.argv:
    setup_requires.extend([
        'setuptools',
        'pytest-runner',
    ])

install_requires = [
    'scipy >= 0.12.1',
    'numpy >= 1.16',
    'astropy >= 1.1.1',
    'configparser',
    'tqdm >= 4.0',
    'pandas >= 2.0.0',
    'tables > 3.5.0',
    'h5py >= 1.3',
    'schwimmbad >= 0.3.1',
    'matplotlib >= 2.0.0',
    'importlib-metadata < 5.0'
]
tests_require = [
    'pytest'
]
extras_require = {
    "doc": [
        "sphinx >= 1.6.1",
        "numpydoc >= 0.8.0",
        "sphinx-bootstrap-theme >= 0.6",
        "sphinxcontrib-programoutput",
        "sphinx-automodapi",
        "ipython",
        "sphinx_rtd_theme",
    ],
}

try:
    from numpy.distutils.core import setup, Extension
except ImportError:
    raise ImportError("Building fortran extensions requires numpy.")
# fortran compile
wrapper = Extension(
    "cosmic._evolvebin",
    sources=[
        "cosmic/src/comenv.f",
        "cosmic/src/corerd.f",
        "cosmic/src/deltat.f",
        "cosmic/src/dgcore.f",
        "cosmic/src/evolv2.f",
        "cosmic/src/gntage.f",
        "cosmic/src/hrdiag.f",
        "cosmic/src/instar.f",
        "cosmic/src/kick.f",
        "cosmic/src/mix.f",
        "cosmic/src/mlwind.f",
        "cosmic/src/mrenv.f",
        "cosmic/src/ran3.f",
        "cosmic/src/rl.f",
        "cosmic/src/star.f",
        "cosmic/src/zcnsts.f",
        "cosmic/src/zfuncs.f",
        "cosmic/src/concatkstars.f",
        "cosmic/src/comprad.f",
        "cosmic/src/bpp_array.f",
        "cosmic/src/checkstate.f",
        "cosmic/src/assign_remnant.f",
        "cosmic/src/hrdiag_remnant.f",

#    ], extra_compile_args = ["-g","-O0"], extra_f77_compile_args=["-O0"], extra_f90_compile_args=["-O0"])
])
# -- run setup ----------------------------------------------------------------

packagenames = find_packages()
scripts = glob.glob(os.path.join("bin", "*"))

setup(name=DISTNAME,
      provides=[PACKAGENAME],
      version=verstr,
      description="Compact Object Synthesis and Monte Carlo Investigation Code",
      long_description=long_description,
      long_description_content_type='text/markdown',
      ext_modules=[wrapper],
      author=AUTHOR,
      author_email=AUTHOR_EMAIL,
      license=LICENSE,
      packages=packagenames,
      include_package_data=True,
      cmdclass=cmdclass,
      url='https://github.com/COSMIC-PopSynth/COSMIC',
      scripts=scripts,
      setup_requires=setup_requires,
      install_requires=install_requires,
      tests_require=tests_require,
      extras_require=extras_require,
      python_requires='>3.6, <4',
      classifiers=[
          'Development Status :: 4 - Beta',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: 3.10',
          'Intended Audience :: Science/Research',
          'Intended Audience :: End Users/Desktop',
          'Intended Audience :: Science/Research',
          'Natural Language :: English',
          'Topic :: Scientific/Engineering',
          'Topic :: Scientific/Engineering :: Astronomy',
          'Topic :: Scientific/Engineering :: Physics',
          'Operating System :: POSIX',
          'Operating System :: Unix',
          'Operating System :: MacOS',
          'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
                  ],
      )
